"""Data models for Whoop API v2 using Pydantic v2.

Copyright (c) 2024 Felix Geilert
"""

from datetime import datetime
from enum import Enum
from uuid import UUID

import pandas as pd
from pydantic import BaseModel, ConfigDict, Field, computed_field


class ScoreState(str, Enum):
    """Enumeration of possible score states."""

    SCORED = "SCORED"
    PENDING_SCORE = "PENDING_SCORE"
    UNSCORABLE = "UNSCORABLE"


class BaseWhoopModel(BaseModel):
    """Base model with common configuration."""

    model_config = ConfigDict(
        from_attributes=True,
        populate_by_name=True,
        use_enum_values=True,
        str_strip_whitespace=True,
    )


# User Models
class UserBasicProfile(BaseWhoopModel):
    """Basic user profile information."""

    user_id: int
    email: str
    first_name: str
    last_name: str


class UserBodyMeasurement(BaseWhoopModel):
    """User body measurements."""

    height_meter: float
    weight_kilogram: float
    max_heart_rate: int


# Cycle Models
class CycleScore(BaseWhoopModel):
    """Cycle scoring metrics."""

    strain: float
    kilojoule: float
    average_heart_rate: int
    max_heart_rate: int

    @computed_field  # type: ignore[misc]
    def calories(self) -> float:
        """Convert kilojoules to calories."""
        return self.kilojoule / 4.184


class Cycle(BaseWhoopModel):
    """Physiological cycle data."""

    id: int
    user_id: int
    created_at: datetime
    updated_at: datetime
    start: datetime
    end: datetime | None = None
    timezone_offset: str
    score_state: ScoreState
    score: CycleScore | None = None

    @computed_field  # type: ignore[misc]
    def duration_hours(self) -> float | None:
        """Calculate cycle duration in hours."""
        if self.end:
            return (self.end - self.start).total_seconds() / 3600
        return None

    @computed_field  # type: ignore[misc]
    def is_complete(self) -> bool:
        """Check if the cycle is complete."""
        return self.end is not None


# Sleep Models
class SleepStageSummary(BaseWhoopModel):
    """Summary of sleep stages."""

    total_in_bed_time_milli: int
    total_awake_time_milli: int
    total_no_data_time_milli: int
    total_light_sleep_time_milli: int
    total_slow_wave_sleep_time_milli: int
    total_rem_sleep_time_milli: int
    sleep_cycle_count: int
    disturbance_count: int

    @computed_field  # type: ignore[misc]
    def total_sleep_time_milli(self) -> int:
        """Calculate total sleep time."""
        return (
            self.total_light_sleep_time_milli + self.total_slow_wave_sleep_time_milli + self.total_rem_sleep_time_milli
        )

    @computed_field  # type: ignore[misc]
    @property
    def sleep_efficiency_percentage(self) -> float:
        """Calculate sleep efficiency as percentage."""
        if self.total_in_bed_time_milli == 0:
            return 0.0
        # Access the method result directly
        total_sleep = self.total_sleep_time_milli
        return (total_sleep / self.total_in_bed_time_milli) * 100


class SleepNeeded(BaseWhoopModel):
    """Breakdown of sleep need."""

    baseline_milli: int
    need_from_sleep_debt_milli: int
    need_from_recent_strain_milli: int
    need_from_recent_nap_milli: int

    @computed_field  # type: ignore[misc]
    def total_need_milli(self) -> int:
        """Calculate total sleep need."""
        return (
            self.baseline_milli
            + self.need_from_sleep_debt_milli
            + self.need_from_recent_strain_milli
            + self.need_from_recent_nap_milli
        )


class SleepScore(BaseWhoopModel):
    """Sleep scoring metrics."""

    stage_summary: SleepStageSummary
    sleep_needed: SleepNeeded
    respiratory_rate: float | None = None
    sleep_performance_percentage: float | None = None
    sleep_consistency_percentage: float | None = None
    sleep_efficiency_percentage: float | None = None


class Sleep(BaseWhoopModel):
    """Sleep activity data."""

    id: UUID
    v1_id: int | None = Field(None, description="Legacy v1 ID")
    user_id: int
    created_at: datetime
    updated_at: datetime
    start: datetime
    end: datetime
    timezone_offset: str
    nap: bool
    score_state: ScoreState
    score: SleepScore | None = None

    @computed_field  # type: ignore[misc]
    def duration_hours(self) -> float:
        """Calculate sleep duration in hours."""
        return (self.end - self.start).total_seconds() / 3600


# Recovery Models
class RecoveryScore(BaseWhoopModel):
    """Recovery scoring metrics."""

    user_calibrating: bool
    recovery_score: float
    resting_heart_rate: float
    hrv_rmssd_milli: float
    spo2_percentage: float | None = None
    skin_temp_celsius: float | None = None


class Recovery(BaseWhoopModel):
    """Recovery data."""

    cycle_id: int
    sleep_id: UUID
    user_id: int
    created_at: datetime
    updated_at: datetime
    score_state: ScoreState
    score: RecoveryScore | None = None


# Workout Models
class ZoneDurations(BaseWhoopModel):
    """Time spent in each heart rate zone."""

    zone_zero_milli: int
    zone_one_milli: int
    zone_two_milli: int
    zone_three_milli: int
    zone_four_milli: int
    zone_five_milli: int

    @computed_field  # type: ignore[misc]
    def total_duration_milli(self) -> int:
        """Calculate total workout duration."""
        return (
            self.zone_zero_milli
            + self.zone_one_milli
            + self.zone_two_milli
            + self.zone_three_milli
            + self.zone_four_milli
            + self.zone_five_milli
        )

    def to_dict_percentage(self) -> dict[str, float]:
        """Convert zone durations to percentages."""
        # Access the computed field value directly
        total = self.total_duration_milli
        if total == 0:
            return {f"zone_{i}_percentage": 0.0 for i in range(6)}

        return {
            "zone_zero_percentage": (self.zone_zero_milli / total) * 100,
            "zone_one_percentage": (self.zone_one_milli / total) * 100,
            "zone_two_percentage": (self.zone_two_milli / total) * 100,
            "zone_three_percentage": (self.zone_three_milli / total) * 100,
            "zone_four_percentage": (self.zone_four_milli / total) * 100,
            "zone_five_percentage": (self.zone_five_milli / total) * 100,
        }


class WorkoutScore(BaseWhoopModel):
    """Workout scoring metrics."""

    strain: float
    average_heart_rate: int
    max_heart_rate: int
    kilojoule: float
    percent_recorded: float
    distance_meter: float | None = None
    altitude_gain_meter: float | None = None
    altitude_change_meter: float | None = None
    zone_durations: ZoneDurations

    @computed_field  # type: ignore[misc]
    def calories(self) -> float:
        """Convert kilojoules to calories."""
        return self.kilojoule / 4.184


class WorkoutV2(BaseWhoopModel):
    """Workout activity data."""

    id: UUID
    v1_id: int | None = Field(None, description="Legacy v1 ID")
    sport_id: int | None = Field(None, description="Legacy sport ID")
    user_id: int
    created_at: datetime
    updated_at: datetime
    start: datetime
    end: datetime
    timezone_offset: str
    sport_name: str
    score_state: ScoreState
    score: WorkoutScore | None = None

    @computed_field  # type: ignore[misc]
    def duration_hours(self) -> float:
        """Calculate workout duration in hours."""
        return (self.end - self.start).total_seconds() / 3600


# Collection Response Models
class PaginatedResponse(BaseWhoopModel):
    """Base class for paginated responses."""

    next_token: str | None = None

    @computed_field  # type: ignore[misc]
    def has_more(self) -> bool:
        """Check if there are more pages."""
        return self.next_token is not None


class PaginatedCycleResponse(PaginatedResponse):
    """Paginated response for cycles."""

    records: list[Cycle]


class PaginatedSleepResponse(PaginatedResponse):
    """Paginated response for sleep activities."""

    records: list[Sleep]


class RecoveryCollection(PaginatedResponse):
    """Paginated response for recoveries."""

    records: list[Recovery]


class WorkoutCollection(PaginatedResponse):
    """Paginated response for workouts."""

    records: list[WorkoutV2]


# Sport ID mapping (for backward compatibility)
SPORT_IDS = {
    1: "running",
    2: "cycling",
    3: "swimming",
    4: "strength_training",
    5: "yoga",
    6: "pilates",
    7: "spinning",
    8: "elliptical",
    9: "stairmaster",
    10: "rowing",
    11: "bootcamp",
    12: "boxing",
    13: "functional_fitness",
    14: "dance",
    15: "barre",
    16: "hiking",
    17: "golf",
    18: "tennis",
    19: "squash",
    20: "basketball",
    21: "soccer",
    22: "football",
    23: "hockey",
    24: "lacrosse",
    25: "rugby",
    26: "volleyball",
    27: "softball",
    28: "baseball",
    29: "skiing",
    30: "snowboarding",
    31: "crossfit",
    32: "rock_climbing",
    33: "surfing",
    34: "paddleboarding",
    35: "kayaking",
    36: "ice_bath",
    37: "meditation",
    38: "other",
    39: "walking",
    40: "water_polo",
    41: "wrestling",
    42: "martial_arts",
    43: "gymnastics",
    44: "track_field",
    45: "roller_skating",
    46: "ice_skating",
    47: "crew",
    48: "cricket",
    49: "pickleball",
    50: "inline_skating",
    51: "skateboarding",
    52: "badminton",
    53: "table_tennis",
    54: "climbing",
    55: "assault_bike",
    56: "parkour",
    57: "jiu_jitsu",
    58: "rotational",
    59: "jump_rope",
    60: "skiing_nordic",
    61: "manual_labor",
    62: "ultimate",
    63: "motocross",
    64: "stairs",
    65: "sailing",
    66: "hyrox",
    67: "canoeing",
    68: "field_hockey",
    69: "ruck",
    70: "triathlon",
    71: "paddle_tennis",
    72: "stretching",
    73: "gaelic_football",
    74: "hurling_camogie",
    75: "australian_football",
    76: "sauna",
    77: "calisthenics",
    78: "massage_rolling",
    79: "breathwork",
    80: "wheelchair_push",
    81: "waterski_wakeboard",
    82: "wind_kite_surfing",
    83: "bowling",
    84: "duathlon",
    85: "biathlon",
    86: "netball",
    87: "horseback_riding",
    88: "scuba_diving",
    89: "freediving",
    90: "boxing_bag",
    91: "tennis_paddle",
    92: "hunting",
    93: "fishing",
    94: "shooting",
    95: "archery",
    96: "rafting",
    97: "kickboxing",
    98: "muay_thai",
    99: "weightlifting",
    100: "powerlifting",
    101: "spartan",
    102: "tough_mudder",
    103: "flag_football",
    104: "racquetball",
    105: "fencing",
    106: "trampoline",
    107: "nordic_walking",
    108: "prehab_rehab",
    109: "sled",
    110: "tubing",
    111: "rodeo",
    112: "wrestling_entertainment",
}


# Helper functions for DataFrame conversion
def models_to_dataframe(models: list[BaseWhoopModel]) -> pd.DataFrame:
    """
    Convert a list of Pydantic models to a pandas DataFrame.

    This function takes a list of Whoop data models and converts them to a
    DataFrame with normalized/flattened structure. Datetime columns are
    automatically converted to pandas datetime types.

    Args:
        models: List of BaseWhoopModel instances to convert

    Returns:
        DataFrame with flattened structure from the models, or empty DataFrame if no models
    """
    if not models:
        return pd.DataFrame()

    # Convert models to dictionaries
    data = [model.model_dump() for model in models]

    # Create DataFrame
    df = pd.json_normalize(data)

    # Convert datetime columns
    datetime_columns = [col for col in df.columns if col.endswith("_at") or col in ["start", "end"]]
    for col in datetime_columns:
        if col in df.columns:
            df[col] = pd.to_datetime(df[col])

    return df



"""Holds data models for the whoop api.

Copyright (c) 2022 Felix Geilert
"""

from abc import abstractmethod
from datetime import datetime, timedelta
from typing import Any

import time_helper as th
from pydantic import BaseModel


class UserProfile(BaseModel):
    """Represents a user profile."""

    user_id: int
    email: str
    first_name: str
    last_name: str


class UserMeasurements(BaseModel):
    """Stores user measurements."""

    height_meter: float
    weight_kilogram: float
    max_heart_rate: int


class UserData(BaseModel):
    """Base Data that is used for most user data objects"""

    user_id: int
    created_at: datetime
    updated_at: datetime
    timezone_offset: str | None = None
    score_state: str

    @classmethod
    @abstractmethod
    def _dict_parse(cls, data: dict) -> dict[str, Any]:
        return data

    @classmethod
    def from_dict(cls, data: dict, correct_offset: bool = False) -> "UserData":
        # generate timedelta from timezone offset
        td = timedelta(days=0)
        if correct_offset and "timezone_offset" in data:
            to_str = data["timezone_offset"]

            # check if negative
            negative = to_str[0] == "-"
            if to_str[0] in ["+", "-"]:
                to_str = to_str[1:]

            # parse hours and minutes to int
            hours, minutes = to_str.split(":")
            hours = int(hours)
            minutes = int(minutes)

            # generate timedelta
            td = timedelta(hours=hours, minutes=minutes)
            if negative:
                td *= -1

        # update the datetimes
        for dt in ["created_at", "updated_at", "start", "end"]:
            if dt in data and data[dt] is not None:
                date = datetime.strptime(data[dt], "%Y-%m-%dT%H:%M:%S.%fZ")
                data[dt] = th.any_to_datetime(date) + td
        data = cls._dict_parse(data)

        return cls(**data)


class UserCycleScore(BaseModel):
    strain: float
    kilojoule: float
    average_heart_rate: int
    max_heart_rate: int

    @property
    def calories(self) -> float:
        return self.kilojoule / 4.184


class UserCycle(UserData):
    """Stores data of the user cycle."""

    id: int
    start: datetime
    end: datetime | None = None
    score: UserCycleScore | None = None

    @classmethod
    def _dict_parse(cls, data: dict) -> dict[str, Any]:
        if "score" in data and data["score"] is not None:
            data["score"] = UserCycleScore(**data["score"])

        return data


class UserSleepStages(BaseModel):
    """Stores the sleep stages."""

    total_in_bed_time_milli: int
    total_awake_time_milli: int
    total_no_data_time_milli: int
    total_light_sleep_time_milli: int
    total_slow_wave_sleep_time_milli: int
    total_rem_sleep_time_milli: int
    sleep_cycle_count: int
    disturbance_count: int


class UserSleepNeed(BaseModel):
    """Stores the sleep need."""

    baseline_milli: int
    need_from_sleep_debt_milli: int
    need_from_recent_strain_milli: int
    need_from_recent_nap_milli: int


class UserSleepScore(BaseModel):
    """Stores the score of the user sleep."""

    stage_summary: UserSleepStages | None = None
    sleep_needed: UserSleepNeed | None = None
    respiratory_rate: float | None = None
    sleep_performance_percentage: float | None = None
    sleep_consistency_percentage: float | None = None
    sleep_efficiency_percentage: float | None = None


class UserSleep(UserData):
    """Stores data for user sleep."""

    id: int
    nap: bool
    start: datetime
    end: datetime | None = None
    score: UserSleepScore | None = None

    @classmethod
    def _dict_parse(cls, data: dict) -> dict[str, Any]:
        if "score" in data and data["score"] is not None:
            score_dict = data["score"]
            if "stage_summary" in score_dict and score_dict["stage_summary"] is not None:
                score_dict["stage_summary"] = UserSleepStages(**score_dict["stage_summary"])
            else:
                score_dict["stage_summary"] = None
            if "sleep_needed" in score_dict and score_dict["sleep_needed"] is not None:
                score_dict["sleep_needed"] = UserSleepNeed(**score_dict["sleep_needed"])
            else:
                score_dict["sleep_needed"] = None
            data["score"] = UserSleepScore(**score_dict)

        return data


class UserRecoveryScore(BaseModel):
    """Score of users recovery."""

    user_calibrating: bool
    recovery_score: float
    resting_heart_rate: float
    hrv_rmssd_milli: float
    spo2_percentage: float | None = None
    skin_temp_celsius: float | None = None


class UserRecovery(UserData):
    """Stores data for user recovery."""

    cycle_id: int
    sleep_id: int
    score: UserRecoveryScore

    @classmethod
    def _dict_parse(cls, data: dict) -> dict[str, Any]:
        if "score" in data and data["score"] is not None:
            data["score"] = UserRecoveryScore(**data["score"])

        return data


class UserWorkoutZoneDuration(BaseModel):
    """Stores the duration of the workout zone."""

    zone_zero_milli: int
    zone_one_milli: int
    zone_two_milli: int
    zone_three_milli: int
    zone_four_milli: int
    zone_five_milli: int


class UserWorkoutScore(BaseModel):
    """Score of the user workout."""

    strain: float
    average_heart_rate: int
    max_heart_rate: int
    kilojoule: float
    percent_recorded: float
    distance_meter: float | None = None
    altitude_gain_meter: float | None = None
    altitude_change_meter: float | None = None
    zone_duration: UserWorkoutZoneDuration | None = None


class UserWorkout(UserData):
    """Stores information about a user workout"""

    id: int
    start: datetime
    end: datetime | None = None
    sport_id: int
    score: UserWorkoutScore | None = None

    @classmethod
    def _dict_parse(cls, data: dict) -> dict[str, Any]:
        if "score" in data and data["score"] is not None:
            score_dict = data["score"]
            if "zone_duration" in score_dict and score_dict["zone_duration"] is not None:
                score_dict["zone_duration"] = UserWorkoutZoneDuration(**score_dict["zone_duration"])
            data["score"] = UserWorkoutScore(**score_dict)

        return data


SPORT_IDS = {
    -1: "Activity",
    0: "Running",
    1: "Cycling",
    16: "Baseball",
    17: "Basketball",
    18: "Rowing",
    19: "Fencing",
    20: "Field Hockey",
    21: "Football",
    22: "Golf",
    24: "Ice Hockey",
    25: "Lacrosse",
    27: "Rugby",
    28: "Sailing",
    29: "Skiing",
    30: "Soccer",
    31: "Softball",
    32: "Squash",
    33: "Swimming",
    34: "Tennis",
    35: "Track & Field",
    36: "Volleyball",
    37: "Water Polo",
    38: "Wrestling",
    39: "Boxing",
    42: "Dance",
    43: "Pilates",
    44: "Yoga",
    45: "Weightlifting",
    47: "Cross Country Skiing",
    48: "Functional Fitness",
    49: "Duathlon",
    51: "Gymnastics",
    52: "Hiking/Rucking",
    53: "Horseback Riding",
    55: "Kayaking",
    56: "Martial Arts",
    57: "Mountain Biking",
    59: "Powerlifting",
    60: "Rock Climbing",
    61: "Paddleboarding",
    62: "Triathlon",
    63: "Walking",
    64: "Surfing",
    65: "Elliptical",
    66: "Stairmaster",
    70: "Meditation",
    71: "Other",
    73: "Diving",
    74: "Operations - Tactical",
    75: "Operations - Medical",
    76: "Operations - Flying",
    77: "Operations - Water",
    82: "Ultimate",
    83: "Climber",
    84: "Jumping Rope",
    85: "Australian Football",
    86: "Skateboarding",
    87: "Coaching",
    88: "Ice Bath",
    89: "Commuting",
    90: "Gaming",
    91: "Snowboarding",
    92: "Motocross",
    93: "Caddying",
    94: "Obstacle Course Racing",
    95: "Motor Racing",
    96: "HIIT",
    97: "Spin",
    98: "Jiu Jitsu",
    99: "Manual Labor",
    100: "Cricket",
    101: "Pickleball",
    102: "Inline Skating",
    103: "Box Fitness",
    104: "Spikeball",
    105: "Wheelchair Pushing",
    106: "Paddle Tennis",
    107: "Barre",
    108: "Stage Performance",
    109: "High Stress Work",
    110: "Parkour",
    111: "Gaelic Football",
    112: "Hurling/Camogie",
    113: "Circus Arts",
    121: "Massage Therapy",
    125: "Watching Sports",
    126: "Assault Bike",
    127: "Kickboxing",
    128: "Stretching",
    230: "Table Tennis",
    231: "Badminton",
    232: "Netball",
    233: "Sauna",
    234: "Disc Golf",
    235: "Yard Work",
    236: "Air Compression",
    237: "Percussive Massage",
    238: "Paintball",
    239: "Ice Skating",
    240: "Handball",
}



"""Pytest configuration for whoopy tests.

Copyright (c) 2024 Felix Geilert
"""

import asyncio

import pytest


@pytest.fixture(scope="session")
def event_loop():
    """Create an instance of the default event loop for the test session."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


# Configure pytest-asyncio
pytest_plugins = ["pytest_asyncio"]



__version__ = "0.3.0"

import logging
from typing import Any

# Import v1 for backward compatibility
try:
    from . import handlers as _handlers_v1  # noqa: F401
    from . import models as _models_v1  # noqa: F401
    from .client_v1 import API_VERSION as API_VERSION_V1
    from .client_v1 import WhoopClient as WhoopClientV1
    from .models.models_v1 import SPORT_IDS as SPORT_IDS_V1

    v1_available = True
except Exception as ex:
    logging.error(f"Error importing whoopy v1: {ex}")
    v1_available = False
    WhoopClientV1 = None  # type: ignore[misc,assignment]
    API_VERSION_V1 = "1"
    SPORT_IDS_V1 = {}

# Import v2 as the new default
WhoopClient: Any
WhoopClientV2: Any | None = None
WhoopClientV2Sync: Any | None = None
API_VERSION: str
SPORT_IDS: dict[int, str]

try:
    from .client_v2 import WhoopClientV2
    from .models.models_v2 import SPORT_IDS
    from .sync_wrapper import WhoopClientV2Sync

    # Make v2 sync wrapper the default
    WhoopClient = WhoopClientV2Sync
    API_VERSION = "2"
    v2_available = True

except Exception as ex:
    logging.error(f"Error importing whoopy v2: {ex}")
    v2_available = False
    # Fall back to v1 if v2 fails and v1 is available
    if v1_available:
        WhoopClient = WhoopClientV1
        API_VERSION = API_VERSION_V1
        SPORT_IDS = SPORT_IDS_V1
    else:
        # Neither v1 nor v2 available
        raise ImportError("Unable to import any version of WhoopClient") from ex

# Export all available clients
__all__ = [
    "API_VERSION",  # API version (v2 by default)
    "SPORT_IDS",  # Sport ID mapping (v2 by default)
    "WhoopClient",  # Default (v2 sync)
    "WhoopClientV1",  # Legacy v1
    "WhoopClientV2",  # Async v2
    "WhoopClientV2Sync",  # Explicit sync v2
    "__version__",
]

try:
    # import other versions
    from . import client_v1 as _client_v1  # noqa: F401
    from . import client_vu7 as _client_vu7  # noqa: F401
except Exception as ex:
    logging.error(f"Not all dependencies installed: {ex}")



"""Resource handlers for Whoop API v2.

Copyright (c) 2024 Felix Geilert
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from uuid import UUID

from whoopy.exceptions import ResourceNotFoundError
from whoopy.models import models_v2 as models

if TYPE_CHECKING:
    from whoopy.client_v2 import WhoopClientV2

from .base_v2 import BaseHandler, CollectionHandler, CombinedHandler


class UserHandler(BaseHandler):
    """Handler for user-related endpoints."""

    async def get_profile(self) -> models.UserBasicProfile:
        """Get the authenticated user's basic profile."""
        data = await self._get("user/profile/basic")
        return models.UserBasicProfile(**data)

    async def get_body_measurements(self) -> models.UserBodyMeasurement:
        """Get the authenticated user's body measurements."""
        data = await self._get("user/measurement/body")
        return models.UserBodyMeasurement(**data)


class CycleHandler(CombinedHandler[models.Cycle]):
    """Handler for cycle endpoints."""

    def __init__(self, client: WhoopClientV2):
        """
        Initialize CycleHandler.

        Args:
            client: WhoopClientV2 instance for making API requests
        """
        super().__init__(
            client=client,
            resource_path="cycle",
            collection_path="cycle",
            model_class=models.Cycle,
            response_class=models.PaginatedCycleResponse,  # type: ignore[arg-type]
        )

    async def get_sleep(self, cycle_id: int) -> models.Sleep:
        """
        Get the sleep for a specific cycle.

        Args:
            cycle_id: The cycle ID

        Returns:
            Sleep data for the cycle

        Raises:
            ResourceNotFoundError: If cycle or sleep not found
        """
        try:
            data = await self._get(f"cycle/{cycle_id}/sleep")
            return models.Sleep(**data)
        except Exception as e:
            if "404" in str(e):
                raise ResourceNotFoundError(resource_type="Sleep for Cycle", resource_id=str(cycle_id)) from e
            raise


class SleepHandler(CombinedHandler[models.Sleep]):
    """Handler for sleep endpoints."""

    def __init__(self, client: WhoopClientV2):
        """
        Initialize SleepHandler.

        Args:
            client: WhoopClientV2 instance for making API requests
        """
        super().__init__(
            client=client,
            resource_path="activity/sleep",
            collection_path="activity/sleep",
            model_class=models.Sleep,
            response_class=models.PaginatedSleepResponse,  # type: ignore[arg-type]
        )

    async def get_by_id(self, sleep_id: str | UUID) -> models.Sleep:  # type: ignore[override]
        """
        Get a sleep activity by ID.

        Args:
            sleep_id: The sleep UUID

        Returns:
            Sleep data
        """
        # Convert UUID to string if needed
        if isinstance(sleep_id, UUID):
            sleep_id = str(sleep_id)

        return await super().get_by_id(sleep_id)


class RecoveryHandler(CollectionHandler[models.Recovery]):
    """Handler for recovery endpoints."""

    def __init__(self, client: WhoopClientV2):
        """
        Initialize RecoveryHandler.

        Args:
            client: WhoopClientV2 instance for making API requests
        """
        super().__init__(
            client=client,
            collection_path="activity/recovery",
            model_class=models.Recovery,
            response_class=models.RecoveryCollection,  # type: ignore[arg-type]
        )

    async def get_for_cycle(self, cycle_id: int) -> models.Recovery:
        """
        Get the recovery for a specific cycle.

        Args:
            cycle_id: The cycle ID

        Returns:
            Recovery data for the cycle

        Raises:
            ResourceNotFoundError: If recovery not found
        """
        try:
            data = await self._get(f"activity/recovery/cycle/{cycle_id}/recovery")
            return models.Recovery(**data)
        except Exception as e:
            if "404" in str(e):
                raise ResourceNotFoundError(resource_type="Recovery for Cycle", resource_id=str(cycle_id)) from e
            raise


class WorkoutHandler(CombinedHandler[models.WorkoutV2]):
    """Handler for workout endpoints."""

    def __init__(self, client: WhoopClientV2):
        """
        Initialize WorkoutHandler.

        Args:
            client: WhoopClientV2 instance for making API requests
        """
        super().__init__(
            client=client,
            resource_path="activity/workout",
            collection_path="activity/workout",
            model_class=models.WorkoutV2,
            response_class=models.WorkoutCollection,  # type: ignore[arg-type]
        )

    async def get_by_id(self, workout_id: str | UUID) -> models.WorkoutV2:  # type: ignore[override]
        """
        Get a workout by ID.

        Args:
            workout_id: The workout UUID

        Returns:
            Workout data
        """
        # Convert UUID to string if needed
        if isinstance(workout_id, UUID):
            workout_id = str(workout_id)

        return await super().get_by_id(workout_id)

    async def get_by_sport(
        self,
        sport_name: str,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        limit_per_page: int = 25,
        max_records: int | None = None,
    ) -> list[models.WorkoutV2]:
        """
        Get all workouts for a specific sport.

        Args:
            sport_name: Name of the sport
            start: Start time filter
            end: End time filter
            limit_per_page: Items per page
            max_records: Maximum records to fetch

        Returns:
            List of workouts for the sport
        """
        all_workouts = await self.get_all(start=start, end=end, limit_per_page=limit_per_page, max_records=max_records)

        # Filter by sport name
        sport_name_lower = sport_name.lower()
        return [w for w in all_workouts if w.sport_name.lower() == sport_name_lower]


# Export all handlers
__all__ = [
    "CycleHandler",
    "RecoveryHandler",
    "SleepHandler",
    "UserHandler",
    "WorkoutHandler",
]



"""Whoop API v2 client with async/await support.

This module provides the main client for interacting with the Whoop API v2.
It features modern async/await patterns, automatic retry logic, and proper
error handling.

Follows api docs: https://api.prod.whoop.com/developer/doc/openapi.json

Copyright (c) 2024 Felix Geilert
"""

import json
import logging
import os
from typing import Any

import aiohttp

from .constants import (
    HTTP_BAD_REQUEST,
    HTTP_INTERNAL_SERVER_ERROR,
    HTTP_NOT_FOUND,
    HTTP_OK,
    HTTP_TOO_MANY_REQUESTS,
    HTTP_UNAUTHORIZED,
)
from .exceptions import (
    AuthenticationError,
    ConfigurationError,
    RateLimitError,
    RefreshTokenError,
    ResourceNotFoundError,
    ServerError,
    TokenExpiredError,
    ValidationError,
    WhoopException,
)
from .handlers import handlers_v2 as handlers
from .utils import OAuth2Helper, RequestThrottler, RetryableSession, RetryConfig, TokenInfo

API_VERSION = "2"
API_BASE = "https://api.prod.whoop.com/"


class WhoopClientV2:
    """Async client for Whoop API v2."""

    def __init__(
        self,
        token_info: TokenInfo | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        redirect_uri: str = "http://localhost:1234",
        retry_config: RetryConfig | None = None,
        auto_refresh_token: bool = True,
        logger: logging.Logger | None = None,
        request_delay: float = 0.0,
        max_concurrent_requests: int = 10,
    ):
        """
        Initialize the Whoop API v2 client.

        Args:
            token_info: OAuth2 token information
            client_id: OAuth2 client ID (required for token refresh)
            client_secret: OAuth2 client secret (required for token refresh)
            redirect_uri: OAuth2 redirect URI
            retry_config: Configuration for retry behavior
            auto_refresh_token: Automatically refresh expired tokens
            logger: Logger instance (creates default if None)
            request_delay: Delay in seconds between requests (default: 0)
            max_concurrent_requests: Maximum concurrent requests (default: 10)
        """
        self.token_info = token_info
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.retry_config = retry_config or RetryConfig()
        self.auto_refresh_token = auto_refresh_token
        self.logger = logger or logging.getLogger("whoopy")
        self.request_delay = request_delay
        self.max_concurrent_requests = max_concurrent_requests

        # Session will be created in __aenter__
        self._session: aiohttp.ClientSession | None = None
        self._retry_session: RetryableSession | None = None

        # Request throttler
        self._throttler = RequestThrottler(delay=self.request_delay, max_concurrent=self.max_concurrent_requests)

        # OAuth helper
        if client_id and client_secret:
            self._oauth_helper = OAuth2Helper(
                client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri
            )
        else:
            self._oauth_helper = None  # type: ignore[assignment]

        # API handlers (will be initialized after session is created)
        self._cycles: handlers.CycleHandler | None = None
        self._sleep: handlers.SleepHandler | None = None
        self._recovery: handlers.RecoveryHandler | None = None
        self._workouts: handlers.WorkoutHandler | None = None
        self._user: handlers.UserHandler | None = None

    @property
    def base_url(self) -> str:
        """Get the base URL for API requests."""
        return f"{API_BASE}developer/v{API_VERSION}"

    @property
    def session(self) -> aiohttp.ClientSession:
        """Get the HTTP session."""
        if self._session is None:
            raise RuntimeError("Client session not initialized. Use 'async with' context manager.")
        return self._session

    @property
    def retry_session(self) -> RetryableSession:
        """Get the retry-enabled session."""
        if self._retry_session is None:
            raise RuntimeError("Client session not initialized. Use 'async with' context manager.")
        return self._retry_session

    # Handler properties
    @property
    def cycles(self) -> handlers.CycleHandler:
        """Get the cycles handler."""
        if self._cycles is None:
            raise RuntimeError("Client not initialized. Use 'async with' context manager.")
        return self._cycles

    @property
    def sleep(self) -> handlers.SleepHandler:
        """Get the sleep handler."""
        if self._sleep is None:
            raise RuntimeError("Client not initialized. Use 'async with' context manager.")
        return self._sleep

    @property
    def recovery(self) -> handlers.RecoveryHandler:
        """Get the recovery handler."""
        if self._recovery is None:
            raise RuntimeError("Client not initialized. Use 'async with' context manager.")
        return self._recovery

    @property
    def workouts(self) -> handlers.WorkoutHandler:
        """Get the workouts handler."""
        if self._workouts is None:
            raise RuntimeError("Client not initialized. Use 'async with' context manager.")
        return self._workouts

    @property
    def user(self) -> handlers.UserHandler:
        """Get the user handler."""
        if self._user is None:
            raise RuntimeError("Client not initialized. Use 'async with' context manager.")
        return self._user

    async def __aenter__(self) -> "WhoopClientV2":
        """Enter async context manager."""
        # Check if we have authentication
        if self.token_info is None:
            self.logger.error("No authentication token available")
            raise AuthenticationError(
                "No authentication token available. Please authenticate first using "
                "WhoopClientV2.auth_flow() or provide a valid token."
            )

        # Check if token needs refresh before creating session
        if (
            self.token_info.is_expired
            and self.auto_refresh_token
            and self.token_info.refresh_token
            and self._oauth_helper
        ):
            self.logger.info("Token expired, attempting to refresh")
            try:
                # Create temporary session for token refresh
                async with aiohttp.ClientSession() as temp_session:
                    self.token_info = await self._oauth_helper.refresh_access_token(
                        temp_session, self.token_info.refresh_token
                    )
                self.logger.info("Token refreshed successfully")
            except Exception as e:
                # Log but don't fail - let the first API call handle auth errors
                self.logger.warning(f"Failed to refresh expired token: {e}")

        # Create session with proper headers
        headers = {
            "User-Agent": "Whoopy/0.3.0 (Python Whoop API Client)",
            "Accept": "application/json",
        }

        if self.token_info:
            headers["Authorization"] = f"{self.token_info.token_type} {self.token_info.access_token}"

        self._session = aiohttp.ClientSession(headers=headers)
        self._retry_session = RetryableSession(self._session, self.retry_config, self.check_response)

        # Initialize handlers
        self._cycles = handlers.CycleHandler(self)
        self._sleep = handlers.SleepHandler(self)
        self._recovery = handlers.RecoveryHandler(self)
        self._workouts = handlers.WorkoutHandler(self)
        self._user = handlers.UserHandler(self)

        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit async context manager."""
        if self._session:
            await self._session.close()

    async def check_response(self, response: aiohttp.ClientResponse) -> None:
        """
        Check API response for errors.

        Args:
            response: The aiohttp response object

        Raises:
            Various WhoopException subclasses based on status code
        """
        if response.status == HTTP_OK:
            return

        # Try to get error details from response
        try:
            error_data = await response.json()
        except Exception:
            error_data = await response.text()

        # Map status codes to exceptions
        if response.status == HTTP_BAD_REQUEST:
            raise ValidationError(
                message="Bad request",
                validation_errors=error_data if isinstance(error_data, dict) else None,
                details={"status": HTTP_BAD_REQUEST, "response": error_data},
            )
        if response.status == HTTP_UNAUTHORIZED:
            if self.auto_refresh_token and self.token_info and self.token_info.refresh_token:
                # Token might be expired, try to refresh
                raise TokenExpiredError(details={"status": HTTP_UNAUTHORIZED, "response": error_data})
            raise AuthenticationError(details={"status": HTTP_UNAUTHORIZED, "response": error_data})
        if response.status == HTTP_NOT_FOUND:
            raise ResourceNotFoundError(
                resource_type="Resource", details={"status": HTTP_NOT_FOUND, "response": error_data}
            )
        if response.status == HTTP_TOO_MANY_REQUESTS:
            retry_after = response.headers.get("Retry-After")
            self.logger.warning(f"Rate limit hit. Retry-After: {retry_after}")
            # Adjust throttler delay when we hit rate limits
            self._throttler.adjust_delay(factor=2.0)
            raise RateLimitError(
                retry_after=int(retry_after) if retry_after else None,
                details={"status": HTTP_TOO_MANY_REQUESTS, "response": error_data},
            )
        if response.status >= HTTP_INTERNAL_SERVER_ERROR:
            raise ServerError(
                status_code=response.status,
                message=f"Server error: {response.status}",
                details={"status": response.status, "response": error_data},
            )
        raise WhoopException(
            f"Unexpected status code: {response.status}", details={"status": response.status, "response": error_data}
        )

    async def refresh_token(self) -> None:
        """
        Refresh the access token using the refresh token.

        This method uses the stored refresh token to obtain a new access token
        from the Whoop OAuth2 endpoint. The token info is automatically updated.

        Raises:
            ConfigurationError: If OAuth helper not configured (missing client_id/secret)
            RefreshTokenError: If refresh fails or no refresh token available
        """
        if not self._oauth_helper:
            raise ConfigurationError("Cannot refresh token without client_id and client_secret")

        if not self.token_info or not self.token_info.refresh_token:
            raise RefreshTokenError("No refresh token available")

        try:
            self.token_info = await self._oauth_helper.refresh_access_token(self.session, self.token_info.refresh_token)

            # Update session headers with new token
            if self._session:
                self._session.headers["Authorization"] = f"{self.token_info.token_type} {self.token_info.access_token}"

        except Exception as e:
            raise RefreshTokenError(f"Failed to refresh token: {e!s}") from e

    async def request(
        self,
        method: str,
        path: str,
        params: dict[str, Any] | None = None,
        json_data: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> aiohttp.ClientResponse:
        """
        Make an authenticated request to the API.

        This method handles token refresh if needed.

        Args:
            method: HTTP method
            path: API path (relative to base URL)
            params: Query parameters
            json_data: JSON request body
            **kwargs: Additional arguments for aiohttp

        Returns:
            The response object
        """
        url = f"{self.base_url}/{path}"

        # Apply throttling
        async with self._throttler:
            self.logger.debug(f"Request: {method} {url} params={params}")

            # Try the request
            try:
                response = await self.retry_session.request(method, url, params=params, json=json_data, **kwargs)
                self.logger.debug(f"Response: {response.status} for {method} {url}")
                return response  # type: ignore[no-any-return]

            except TokenExpiredError:
                self.logger.info("Token expired during request, refreshing and retrying")
                # Try to refresh token and retry once
                await self.refresh_token()
                response = await self.retry_session.request(method, url, params=params, json=json_data, **kwargs)
                self.logger.debug(f"Retry response: {response.status} for {method} {url}")
                return response  # type: ignore[no-any-return]

    # Authentication methods
    @classmethod
    async def auth_flow(
        cls,
        client_id: str,
        client_secret: str,
        redirect_uri: str = "http://localhost:1234",
        scopes: list[str] | None = None,
        open_browser: bool = True,
    ) -> "WhoopClientV2":
        """
        Perform OAuth2 authorization flow.

        Args:
            client_id: OAuth2 client ID
            client_secret: OAuth2 client secret
            redirect_uri: OAuth2 redirect URI
            scopes: List of scopes to request
            open_browser: Whether to open the authorization URL in browser

        Returns:
            Authenticated WhoopClientV2 instance
        """
        oauth_helper = OAuth2Helper(
            client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri, scopes=scopes
        )

        # Get authorization URL
        auth_url = oauth_helper.get_authorization_url()

        if open_browser:
            print(f"Opening browser to: {auth_url}")
            oauth_helper.open_authorization_url()
        else:
            print(f"Visit this URL to authorize: {auth_url}")

        # Get authorization code from user
        print(f"\nAfter authorization, you'll be redirected to: {redirect_uri}")
        redirect_url = input("Paste the full redirect URL here: ").strip()

        # Extract code from redirect URL
        from urllib.parse import parse_qs, urlparse

        parsed = urlparse(redirect_url)
        code = parse_qs(parsed.query).get("code", [None])[0]

        if not code:
            raise AuthenticationError("No authorization code found in redirect URL")

        # Exchange code for token
        async with aiohttp.ClientSession() as session:
            token_info = await oauth_helper.exchange_code_for_token(session, code)

        return cls(token_info=token_info, client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri)

    @classmethod
    def from_token(
        cls,
        access_token: str,
        expires_in: int = 3600,
        refresh_token: str | None = None,
        scopes: list[str] | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
    ) -> "WhoopClientV2":
        """
        Create client from existing token.

        Args:
            access_token: The access token
            expires_in: Token expiration time in seconds
            refresh_token: Optional refresh token
            scopes: List of token scopes
            client_id: OAuth2 client ID (for refresh)
            client_secret: OAuth2 client secret (for refresh)

        Returns:
            WhoopClientV2 instance
        """
        token_info = TokenInfo(
            access_token=access_token,
            expires_in=expires_in,
            refresh_token=refresh_token,
            scopes=scopes or OAuth2Helper.DEFAULT_SCOPES,
        )

        return cls(token_info=token_info, client_id=client_id, client_secret=client_secret)

    @classmethod
    def from_config(
        cls, config_path: str = "config.json", token_path: str = ".whoop_credentials.json"
    ) -> "WhoopClientV2":
        """
        Create client from configuration files.

        Args:
            config_path: Path to config file with client credentials
            token_path: Path to saved token file

        Returns:
            WhoopClientV2 instance
        """
        # Load config
        if not os.path.exists(config_path):
            raise ConfigurationError(f"Config file not found: {config_path}")

        with open(config_path) as f:
            config_data = json.load(f)

        # Handle both flat and nested config structures
        assert isinstance(config_data, dict)
        config = config_data.get("whoop", config_data)

        client_id = config.get("client_id")
        client_secret = config.get("client_secret")
        redirect_uri = config.get("redirect_uri", "http://localhost:1234")

        if not client_id or not client_secret:
            raise ConfigurationError("client_id and client_secret required in config")

        # Try to load existing token
        oauth_helper = OAuth2Helper(client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri)

        token_info = oauth_helper.load_token(token_path)

        return cls(token_info=token_info, client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri)

    def save_token(self, path: str = ".whoop_credentials.json") -> None:
        """
        Save current token to file.

        Saves the current OAuth2 token information to a JSON file for later use.

        Args:
            path: Path to save the token file (default: ".whoop_credentials.json")

        Raises:
            ValueError: If no token to save
        """
        if not self.token_info:
            raise ValueError("No token to save")

        if self._oauth_helper:
            self._oauth_helper.save_token(self.token_info, path)
        else:
            # Save manually
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w") as f:
                json.dump(self.token_info.to_dict(), f, indent=2)



"""Constants used throughout the whoopy package.

Copyright (c) 2024 Felix Geilert
"""

# HTTP Status Codes
HTTP_OK = 200
HTTP_BAD_REQUEST = 400
HTTP_UNAUTHORIZED = 401
HTTP_NOT_FOUND = 404
HTTP_TOO_MANY_REQUESTS = 429
HTTP_INTERNAL_SERVER_ERROR = 500

# Other constants
DEFAULT_PAGE_SIZE = 25
MIN_PASSWORD_LENGTH = 8
MAX_RETRY_ATTEMPTS = 3



"""OAuth2 authentication helpers for Whoop API v2.

Copyright (c) 2024 Felix Geilert
"""

import json
import os
import uuid
import webbrowser
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, ClassVar
from urllib.parse import urlencode

import aiohttp

from whoopy.constants import HTTP_OK
from whoopy.exceptions import AuthenticationError, RefreshTokenError


@dataclass
class TokenInfo:
    """Container for OAuth2 token information."""

    access_token: str
    expires_in: int
    refresh_token: str | None
    scopes: list[str]
    token_type: str = "Bearer"
    created_at: datetime | None = None

    def __post_init__(self) -> None:
        if self.created_at is None:
            self.created_at = datetime.now(timezone.utc)

    @property
    def is_expired(self) -> bool:
        """Check if the token has expired."""
        if self.created_at is None:
            return True  # Consider expired if no creation time

        current_time = datetime.now(timezone.utc)
        expiry_time = self.created_at + timedelta(seconds=self.expires_in)

        # Add 60 second buffer to account for clock skew
        is_expired = current_time >= expiry_time - timedelta(seconds=60)

        if is_expired:
            time_since_expiry = current_time - expiry_time
            print(f"Token expired {time_since_expiry.total_seconds():.0f} seconds ago")

        return is_expired

    @property
    def expires_at(self) -> datetime:
        """Get the expiration time of the token."""
        if self.created_at is None:
            # Should not happen due to __post_init__, but mypy doesn't know that
            return datetime.now(timezone.utc)
        return self.created_at + timedelta(seconds=self.expires_in)

    @property
    def time_until_expiry(self) -> timedelta:
        """Get time remaining until token expires."""
        return self.expires_at - datetime.now(timezone.utc)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "access_token": self.access_token,
            "expires_in": self.expires_in,
            "refresh_token": self.refresh_token,
            "scopes": self.scopes,
            "token_type": self.token_type,
            "created_at": self.created_at.isoformat() if self.created_at else None,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "TokenInfo":
        """
        Create TokenInfo from dictionary.

        Args:
            data: Dictionary containing token information with keys:
                - access_token: OAuth2 access token (required)
                - expires_in: Token validity duration in seconds (required)
                - refresh_token: Optional refresh token
                - scopes: Optional list of granted scopes
                - token_type: Token type (default: "Bearer")
                - created_at: Optional ISO format creation timestamp

        Returns:
            TokenInfo instance
        """
        created_at = data.get("created_at")
        if created_at:
            created_at = datetime.fromisoformat(created_at)

        return cls(
            access_token=data["access_token"],
            expires_in=data["expires_in"],
            refresh_token=data.get("refresh_token"),
            scopes=data.get("scopes", []),
            token_type=data.get("token_type", "Bearer"),
            created_at=created_at or datetime.now(timezone.utc),
        )


class OAuth2Helper:
    """Helper class for OAuth2 authentication flow."""

    AUTH_URL = "https://api.prod.whoop.com/oauth/oauth2/auth"
    TOKEN_URL = "https://api.prod.whoop.com/oauth/oauth2/token"

    DEFAULT_SCOPES: ClassVar[list[str]] = [
        "offline",
        "read:recovery",
        "read:cycles",
        "read:sleep",
        "read:workout",
        "read:profile",
        "read:body_measurement",
    ]

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: str = "http://localhost:1234",
        scopes: list[str] | None = None,
    ):
        """
        Initialize OAuth2Helper.

        Args:
            client_id: OAuth2 client ID from Whoop developer portal
            client_secret: OAuth2 client secret from Whoop developer portal
            redirect_uri: Redirect URI for OAuth2 flow (default: "http://localhost:1234")
            scopes: List of OAuth2 scopes to request (default: all read scopes)
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.scopes = scopes or self.DEFAULT_SCOPES

    def get_authorization_url(self, state: str | None = None) -> str:
        """
        Generate the authorization URL for the OAuth2 flow.

        Args:
            state: Optional state parameter for CSRF protection

        Returns:
            Authorization URL to redirect user to
        """
        if state is None:
            state = str(uuid.uuid4())

        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "scope": " ".join(self.scopes),
            "state": state,
        }

        return f"{self.AUTH_URL}?{urlencode(params)}"

    def open_authorization_url(self, state: str | None = None) -> str:
        """
        Open the authorization URL in the default browser.

        Args:
            state: Optional state parameter for CSRF protection

        Returns:
            The state parameter used (generated if not provided)
        """
        url = self.get_authorization_url(state)
        webbrowser.open(url)
        return url

    async def exchange_code_for_token(self, session: aiohttp.ClientSession, code: str) -> TokenInfo:
        """Exchange authorization code for access token."""
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_uri": self.redirect_uri,
        }

        async with session.post(self.TOKEN_URL, data=data) as response:
            if response.status != HTTP_OK:
                error_data = await response.text()
                raise AuthenticationError(
                    f"Failed to exchange code for token: {response.status}",
                    details={"status": response.status, "response": error_data},
                )

            token_data = await response.json()

        return TokenInfo(
            access_token=token_data["access_token"],
            expires_in=token_data["expires_in"],
            refresh_token=token_data.get("refresh_token"),
            scopes=token_data.get("scope", "").split() if token_data.get("scope") else self.scopes,
            token_type=token_data.get("token_type", "Bearer"),
        )

    async def refresh_access_token(self, session: aiohttp.ClientSession, refresh_token: str) -> TokenInfo:
        """
        Refresh an expired access token.

        Args:
            session: aiohttp session for making the request
            refresh_token: Refresh token from previous authentication

        Returns:
            TokenInfo containing the new access token and metadata

        Raises:
            RefreshTokenError: If token refresh fails
        """
        data = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }

        async with session.post(self.TOKEN_URL, data=data) as response:
            if response.status != HTTP_OK:
                error_data = await response.text()
                raise RefreshTokenError(
                    f"Failed to refresh token: {response.status}",
                    details={"status": response.status, "response": error_data},
                )

            token_data = await response.json()

        return TokenInfo(
            access_token=token_data["access_token"],
            expires_in=token_data["expires_in"],
            refresh_token=token_data.get("refresh_token", refresh_token),
            scopes=token_data.get("scope", "").split() if token_data.get("scope") else self.scopes,
            token_type=token_data.get("token_type", "Bearer"),
        )

    def save_token(self, token: TokenInfo, path: str = ".whoop_credentials.json") -> None:
        """
        Save token to file.

        Args:
            token: TokenInfo to save
            path: Path to save the token file (default: ".whoop_credentials.json")
        """
        # Create directory if it doesn't exist
        dir_path = os.path.dirname(path)
        if dir_path and not os.path.exists(dir_path):
            os.makedirs(dir_path, exist_ok=True)

        with open(path, "w") as f:
            json.dump(token.to_dict(), f, indent=2)

    def load_token(self, path: str = ".whoop_credentials.json") -> TokenInfo | None:
        """
        Load token from file.

        Args:
            path: Path to load the token file from (default: ".whoop_credentials.json")

        Returns:
            TokenInfo if file exists and is valid, None otherwise
        """
        if not os.path.exists(path):
            return None

        try:
            with open(path) as f:
                data = json.load(f)
            return TokenInfo.from_dict(data)
        except (json.JSONDecodeError, KeyError, ValueError):
            return None



"""Some basic unit tests for the client

Copyright 2022 (C) Felix Geilert
"""

from whoopy import WhoopClientV1


def test_client_url():
    """Test that the client URL is correct"""
    url, state = WhoopClientV1.auth_url("1234", "5678", "http://localhost:5000")
    assert url == (
        "https://api.prod.whoop.com/oauth/oauth2/auth"
        "?scope=offline read:recovery read:cycles read:sleep read:workout read:profile read:body_measurement"
        f"&client_id=1234&client_secret=5678&state={state}"
        "&redirect_uri=http://localhost:5000&response_type=code"
    )



"""Retry logic with exponential backoff for API requests.

Copyright (c) 2024 Felix Geilert
"""

import asyncio
import logging
import random
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from functools import wraps
from typing import Any, TypeVar, cast

from whoopy.exceptions import RateLimitError, ServerError

T = TypeVar("T")


@dataclass
class RetryConfig:
    """Configuration for retry behavior."""

    max_attempts: int = 5  # Increased for better rate limit handling
    base_delay: float = 1.0
    max_delay: float = 60.0
    exponential_base: float = 2.0
    jitter: bool = True
    retry_on: tuple[type[Exception], ...] = (RateLimitError, ServerError)


def calculate_backoff_delay(attempt: int, config: RetryConfig, retry_after: int | None = None) -> float:
    """
    Calculate the delay before the next retry attempt.

    Args:
        attempt: The attempt number (0-based)
        config: Retry configuration parameters
        retry_after: Server-provided retry delay in seconds (overrides backoff)

    Returns:
        Delay in seconds before next retry
    """
    if retry_after is not None:
        # If server provides retry-after, use it (with small jitter)
        delay = float(retry_after)
        if config.jitter:
            delay += random.uniform(0, 1)
        return min(delay, config.max_delay)

    # Exponential backoff calculation
    delay = config.base_delay * (config.exponential_base**attempt)

    # Add jitter to prevent thundering herd
    if config.jitter:
        delay *= random.uniform(0.8, 1.2)

    return min(delay, config.max_delay)


def retry_with_backoff(
    config: RetryConfig | None = None,
) -> Callable[[Callable[..., Awaitable[T]]], Callable[..., Awaitable[T]]]:
    """
    Decorator for adding retry logic to async functions.

    Args:
        config: Retry configuration (uses defaults if None)

    Returns:
        Decorator function that adds retry logic
    """
    if config is None:
        config = RetryConfig()

    def decorator(func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> T:
            last_exception = None

            for attempt in range(config.max_attempts):
                try:
                    return await func(*args, **kwargs)
                except config.retry_on as e:
                    last_exception = e

                    # Check if this is the last attempt
                    if attempt == config.max_attempts - 1:
                        raise

                    # Calculate delay
                    retry_after = None
                    if isinstance(e, RateLimitError):
                        retry_after = e.retry_after

                    delay = calculate_backoff_delay(attempt, config, retry_after)

                    # Log retry attempt
                    logger = logging.getLogger("whoopy")
                    logger.warning(
                        f"Retrying after {type(e).__name__}. "
                        f"Attempt {attempt + 2}/{config.max_attempts}. "
                        f"Waiting {delay:.1f}s"
                    )

                    # Sleep before retry
                    await asyncio.sleep(delay)

            # This shouldn't be reached, but just in case
            if last_exception:
                raise last_exception
            raise RuntimeError("Unexpected retry logic error")

        return cast(Callable[..., Awaitable[T]], wrapper)

    return decorator


class RetryableSession:
    """A session wrapper that automatically retries failed requests."""

    def __init__(
        self,
        session: Any,
        retry_config: RetryConfig | None = None,
        check_response_func: Callable[[Any], Awaitable[None]] | None = None,
    ) -> None:
        """
        Initialize RetryableSession.

        Args:
            session: The underlying session (e.g., aiohttp.ClientSession)
            retry_config: Configuration for retry behavior (uses defaults if None)
            check_response_func: Optional function to check response validity
        """
        self.session = session
        self.retry_config = retry_config or RetryConfig()
        self.check_response_func = check_response_func

    async def request(self, method: str, url: str, **kwargs: Any) -> Any:
        """
        Make a request with automatic retry logic.

        Args:
            method: HTTP method (GET, POST, etc.)
            url: URL to request
            **kwargs: Additional arguments passed to the underlying session request

        Returns:
            Response from the underlying session

        Raises:
            Various exceptions based on retry configuration
        """

        @retry_with_backoff(self.retry_config)
        async def _request() -> Any:
            response = await self.session.request(method, url, **kwargs)
            if self.check_response_func:
                await self.check_response_func(response)
            return response

        return await _request()

    async def get(self, url: str, **kwargs: Any) -> Any:
        """
        GET request with retry.

        Args:
            url: URL to request
            **kwargs: Additional arguments passed to the request

        Returns:
            Response from the underlying session
        """
        return await self.request("GET", url, **kwargs)

    async def post(self, url: str, **kwargs: Any) -> Any:
        """
        POST request with retry.

        Args:
            url: URL to request
            **kwargs: Additional arguments passed to the request

        Returns:
            Response from the underlying session
        """
        return await self.request("POST", url, **kwargs)

    async def put(self, url: str, **kwargs: Any) -> Any:
        """
        PUT request with retry.

        Args:
            url: URL to request
            **kwargs: Additional arguments passed to the request

        Returns:
            Response from the underlying session
        """
        return await self.request("PUT", url, **kwargs)

    async def delete(self, url: str, **kwargs: Any) -> Any:
        """
        DELETE request with retry.

        Args:
            url: URL to request
            **kwargs: Additional arguments passed to the request

        Returns:
            Response from the underlying session
        """
        return await self.request("DELETE", url, **kwargs)



"""Unit tests for Whoop API v2 handlers.

Copyright (c) 2024 Felix Geilert
"""

import pytest
from aioresponses import aioresponses

from whoopy.client_v2 import WhoopClientV2
from whoopy.models.models_v2 import ScoreState
from whoopy.utils import TokenInfo


class TestUserHandler:
    """Test UserHandler functionality."""

    @pytest.fixture
    def client(self):
        """Create test client."""
        token_info = TokenInfo(
            access_token="test_token",
            expires_in=3600,
            refresh_token=None,
            scopes=["read:profile", "read:body_measurement"],
        )
        return WhoopClientV2(token_info=token_info)

    @pytest.mark.asyncio
    async def test_get_profile(self, client):
        """Test getting user profile."""
        async with client as whoop:
            with aioresponses() as m:
                profile_data = {"user_id": 12345, "email": "test@example.com", "first_name": "John", "last_name": "Doe"}

                m.get("https://api.prod.whoop.com/developer/v2/user/profile/basic", payload=profile_data)

                profile = await whoop.user.get_profile()

                assert profile.user_id == 12345
                assert profile.email == "test@example.com"
                assert profile.first_name == "John"
                assert profile.last_name == "Doe"

    @pytest.mark.asyncio
    async def test_get_body_measurements(self, client):
        """Test getting body measurements."""
        async with client as whoop:
            with aioresponses() as m:
                measurement_data = {"height_meter": 1.80, "weight_kilogram": 75.5, "max_heart_rate": 190}

                m.get("https://api.prod.whoop.com/developer/v2/user/measurement/body", payload=measurement_data)

                measurements = await whoop.user.get_body_measurements()

                assert measurements.height_meter == 1.80
                assert measurements.weight_kilogram == 75.5
                assert measurements.max_heart_rate == 190


class TestCycleHandler:
    """Test CycleHandler functionality."""

    @pytest.fixture
    def client(self):
        """Create test client."""
        token_info = TokenInfo(
            access_token="test_token", expires_in=3600, refresh_token=None, scopes=["read:cycles", "read:sleep"]
        )
        return WhoopClientV2(token_info=token_info)

    @pytest.mark.asyncio
    async def test_get_cycle_by_id(self, client):
        """Test getting single cycle by ID."""
        async with client as whoop:
            with aioresponses() as m:
                cycle_data = {
                    "id": 12345,
                    "user_id": 67890,
                    "created_at": "2024-01-01T00:00:00Z",
                    "updated_at": "2024-01-01T12:00:00Z",
                    "start": "2024-01-01T00:00:00Z",
                    "end": "2024-01-01T23:59:59Z",
                    "timezone_offset": "-05:00",
                    "score_state": "SCORED",
                    "score": {"strain": 10.5, "kilojoule": 2000.0, "average_heart_rate": 120, "max_heart_rate": 180},
                }

                m.get("https://api.prod.whoop.com/developer/v2/cycle/12345", payload=cycle_data)

                cycle = await whoop.cycles.get_by_id(12345)

                assert cycle.id == 12345
                assert cycle.user_id == 67890
                assert cycle.score_state == ScoreState.SCORED
                assert cycle.score.strain == 10.5
                assert cycle.is_complete is True

    @pytest.mark.asyncio
    async def test_get_cycle_collection(self, client):
        """Test getting cycle collection with pagination."""
        async with client as whoop:
            with aioresponses() as m:
                page1_data = {
                    "records": [
                        {
                            "id": 1,
                            "user_id": 100,
                            "created_at": "2024-01-01T00:00:00Z",
                            "updated_at": "2024-01-01T00:00:00Z",
                            "start": "2024-01-01T00:00:00Z",
                            "timezone_offset": "-05:00",
                            "score_state": "SCORED",
                        },
                        {
                            "id": 2,
                            "user_id": 100,
                            "created_at": "2024-01-02T00:00:00Z",
                            "updated_at": "2024-01-02T00:00:00Z",
                            "start": "2024-01-02T00:00:00Z",
                            "timezone_offset": "-05:00",
                            "score_state": "SCORED",
                        },
                    ],
                    "next_token": "token123",
                }

                page2_data = {
                    "records": [
                        {
                            "id": 3,
                            "user_id": 100,
                            "created_at": "2024-01-03T00:00:00Z",
                            "updated_at": "2024-01-03T00:00:00Z",
                            "start": "2024-01-03T00:00:00Z",
                            "timezone_offset": "-05:00",
                            "score_state": "SCORED",
                        }
                    ],
                    "next_token": None,
                }

                # First page (with limit parameter)
                m.get("https://api.prod.whoop.com/developer/v2/cycle?limit=25", payload=page1_data)

                # Second page (with limit and nextToken)
                m.get("https://api.prod.whoop.com/developer/v2/cycle?limit=25&nextToken=token123", payload=page2_data)

                # Get all cycles
                cycles = await whoop.cycles.get_all()

                assert len(cycles) == 3
                assert cycles[0].id == 1
                assert cycles[1].id == 2
                assert cycles[2].id == 3

    @pytest.mark.asyncio
    async def test_get_sleep_for_cycle(self, client):
        """Test getting sleep for a specific cycle."""
        async with client as whoop:
            with aioresponses() as m:
                sleep_data = {
                    "id": "12345678-1234-5678-1234-567812345678",
                    "user_id": 100,
                    "created_at": "2024-01-01T00:00:00Z",
                    "updated_at": "2024-01-01T08:00:00Z",
                    "start": "2024-01-01T00:00:00Z",
                    "end": "2024-01-01T08:00:00Z",
                    "timezone_offset": "-05:00",
                    "nap": False,
                    "score_state": "SCORED",
                }

                m.get("https://api.prod.whoop.com/developer/v2/cycle/12345/sleep", payload=sleep_data)

                sleep = await whoop.cycles.get_sleep(12345)

                assert str(sleep.id) == "12345678-1234-5678-1234-567812345678"
                assert sleep.nap is False
                assert sleep.duration_hours == 8.0


class TestSleepHandler:
    """Test SleepHandler functionality."""

    @pytest.fixture
    def client(self):
        """Create test client."""
        token_info = TokenInfo(access_token="test_token", expires_in=3600, refresh_token=None, scopes=["read:sleep"])
        return WhoopClientV2(token_info=token_info)

    @pytest.mark.asyncio
    async def test_get_sleep_by_id(self, client):
        """Test getting sleep by UUID."""
        async with client as whoop:
            with aioresponses() as m:
                sleep_id = "87654321-4321-8765-4321-876543218765"
                sleep_data = {
                    "id": sleep_id,
                    "v1_id": 12345,
                    "user_id": 100,
                    "created_at": "2024-01-01T00:00:00Z",
                    "updated_at": "2024-01-01T08:00:00Z",
                    "start": "2024-01-01T00:00:00Z",
                    "end": "2024-01-01T08:00:00Z",
                    "timezone_offset": "-05:00",
                    "nap": False,
                    "score_state": "SCORED",
                    "score": {
                        "stage_summary": {
                            "total_in_bed_time_milli": 28800000,
                            "total_awake_time_milli": 1800000,
                            "total_no_data_time_milli": 0,
                            "total_light_sleep_time_milli": 14400000,
                            "total_slow_wave_sleep_time_milli": 7200000,
                            "total_rem_sleep_time_milli": 5400000,
                            "sleep_cycle_count": 4,
                            "disturbance_count": 3,
                        },
                        "sleep_needed": {
                            "baseline_milli": 28800000,
                            "need_from_sleep_debt_milli": 0,
                            "need_from_recent_strain_milli": 0,
                            "need_from_recent_nap_milli": 0,
                        },
                        "respiratory_rate": 16.5,
                        "sleep_performance_percentage": 95.0,
                        "sleep_consistency_percentage": 90.0,
                        "sleep_efficiency_percentage": 93.75,
                    },
                }

                m.get(f"https://api.prod.whoop.com/developer/v2/activity/sleep/{sleep_id}", payload=sleep_data)

                sleep = await whoop.sleep.get_by_id(sleep_id)

                assert str(sleep.id) == sleep_id
                assert sleep.v1_id == 12345
                assert sleep.score.sleep_performance_percentage == 95.0
                assert sleep.score.stage_summary.sleep_efficiency_percentage == 93.75


class TestRecoveryHandler:
    """Test RecoveryHandler functionality."""

    @pytest.fixture
    def client(self):
        """Create test client."""
        token_info = TokenInfo(access_token="test_token", expires_in=3600, refresh_token=None, scopes=["read:recovery"])
        return WhoopClientV2(token_info=token_info)

    @pytest.mark.asyncio
    async def test_get_recovery_for_cycle(self, client):
        """Test getting recovery for a specific cycle."""
        async with client as whoop:
            with aioresponses() as m:
                recovery_data = {
                    "cycle_id": 12345,
                    "sleep_id": "12345678-1234-5678-1234-567812345678",
                    "user_id": 100,
                    "created_at": "2024-01-01T08:00:00Z",
                    "updated_at": "2024-01-01T08:00:00Z",
                    "score_state": "SCORED",
                    "score": {
                        "user_calibrating": False,
                        "recovery_score": 85.0,
                        "resting_heart_rate": 55.0,
                        "hrv_rmssd_milli": 65.5,
                        "spo2_percentage": 98.5,
                        "skin_temp_celsius": 33.2,
                    },
                }

                m.get(
                    "https://api.prod.whoop.com/developer/v2/activity/recovery/cycle/12345/recovery",
                    payload=recovery_data,
                )

                recovery = await whoop.recovery.get_for_cycle(12345)

                assert recovery.cycle_id == 12345
                assert recovery.score.recovery_score == 85.0
                assert recovery.score.hrv_rmssd_milli == 65.5


class TestWorkoutHandler:
    """Test WorkoutHandler functionality."""

    @pytest.fixture
    def client(self):
        """Create test client."""
        token_info = TokenInfo(access_token="test_token", expires_in=3600, refresh_token=None, scopes=["read:workout"])
        return WhoopClientV2(token_info=token_info)

    @pytest.mark.asyncio
    async def test_get_workout_by_id(self, client):
        """Test getting workout by UUID."""
        async with client as whoop:
            with aioresponses() as m:
                workout_id = "11111111-2222-3333-4444-555555555555"
                workout_data = {
                    "id": workout_id,
                    "v1_id": 54321,
                    "sport_id": 1,
                    "user_id": 100,
                    "created_at": "2024-01-01T10:00:00Z",
                    "updated_at": "2024-01-01T11:00:00Z",
                    "start": "2024-01-01T10:00:00Z",
                    "end": "2024-01-01T11:00:00Z",
                    "timezone_offset": "-05:00",
                    "sport_name": "running",
                    "score_state": "SCORED",
                    "score": {
                        "strain": 8.5,
                        "average_heart_rate": 145,
                        "max_heart_rate": 185,
                        "kilojoule": 1500.0,
                        "percent_recorded": 98.5,
                        "distance_meter": 8000.0,
                        "altitude_gain_meter": 150.0,
                        "altitude_change_meter": 10.0,
                        "zone_durations": {
                            "zone_zero_milli": 300000,
                            "zone_one_milli": 600000,
                            "zone_two_milli": 900000,
                            "zone_three_milli": 900000,
                            "zone_four_milli": 600000,
                            "zone_five_milli": 300000,
                        },
                    },
                }

                m.get(f"https://api.prod.whoop.com/developer/v2/activity/workout/{workout_id}", payload=workout_data)

                workout = await whoop.workouts.get_by_id(workout_id)

                assert str(workout.id) == workout_id
                assert workout.sport_name == "running"
                assert workout.score.strain == 8.5
                assert workout.score.distance_meter == 8000.0
                assert workout.duration_hours == 1.0

    @pytest.mark.asyncio
    async def test_get_workouts_by_sport(self, client):
        """Test filtering workouts by sport."""
        async with client as whoop:
            with aioresponses() as m:
                workouts_data = {
                    "records": [
                        {
                            "id": "11111111-1111-1111-1111-111111111111",
                            "user_id": 100,
                            "created_at": "2024-01-01T10:00:00Z",
                            "updated_at": "2024-01-01T11:00:00Z",
                            "start": "2024-01-01T10:00:00Z",
                            "end": "2024-01-01T11:00:00Z",
                            "timezone_offset": "-05:00",
                            "sport_name": "running",
                            "score_state": "SCORED",
                        },
                        {
                            "id": "22222222-2222-2222-2222-222222222222",
                            "user_id": 100,
                            "created_at": "2024-01-02T10:00:00Z",
                            "updated_at": "2024-01-02T11:00:00Z",
                            "start": "2024-01-02T10:00:00Z",
                            "end": "2024-01-02T11:00:00Z",
                            "timezone_offset": "-05:00",
                            "sport_name": "cycling",
                            "score_state": "SCORED",
                        },
                        {
                            "id": "33333333-3333-3333-3333-333333333333",
                            "user_id": 100,
                            "created_at": "2024-01-03T10:00:00Z",
                            "updated_at": "2024-01-03T11:00:00Z",
                            "start": "2024-01-03T10:00:00Z",
                            "end": "2024-01-03T11:00:00Z",
                            "timezone_offset": "-05:00",
                            "sport_name": "running",
                            "score_state": "SCORED",
                        },
                    ],
                    "next_token": None,
                }

                m.get("https://api.prod.whoop.com/developer/v2/activity/workout?limit=25", payload=workouts_data)

                # Get only running workouts
                running_workouts = await whoop.workouts.get_by_sport("running")

                assert len(running_workouts) == 2
                assert all(w.sport_name == "running" for w in running_workouts)



"""Pagination helpers for Whoop API v2.

Copyright (c) 2024 Felix Geilert
"""

from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from whoopy.constants import DEFAULT_PAGE_SIZE

T = TypeVar("T")


@dataclass
class PaginatedResponse(Generic[T]):
    """Container for paginated API responses."""

    records: list[T]
    next_token: str | None = None

    @property
    def has_more(self) -> bool:
        """Check if there are more pages available."""
        return self.next_token is not None


class PaginationHelper(Generic[T]):
    """Helper class for handling paginated API responses."""

    def __init__(self, fetch_page: Callable[..., Awaitable[PaginatedResponse[T]]], model_class: type[T]):
        """
        Initialize pagination helper.

        Args:
            fetch_page: Async function that fetches a single page
            model_class: Pydantic model class for deserializing records
        """
        self.fetch_page = fetch_page
        self.model_class = model_class

    async def get_page(self, limit: int = 10, next_token: str | None = None, **kwargs: Any) -> PaginatedResponse[T]:
        """
        Fetch a single page of results.

        Args:
            limit: Maximum number of items per page (default: 10, max: 25)
            next_token: Token from previous page for pagination
            **kwargs: Additional parameters passed to fetch_page (e.g., start, end)

        Returns:
            PaginatedResponse containing items and optional next_token
        """
        return await self.fetch_page(limit=limit, next_token=next_token, **kwargs)

    async def get_all(self, limit_per_page: int = 25, max_records: int | None = None, **kwargs: Any) -> list[T]:
        """
        Fetch all records across all pages.

        Args:
            limit_per_page: Number of records per page (max 25)
            max_records: Maximum total records to fetch (None = all)
            **kwargs: Additional parameters to pass to fetch_page

        Returns:
            List of all records
        """
        all_records: list[T] = []
        next_token = None

        while True:
            # Calculate how many records to fetch this page
            if max_records is not None:
                remaining = max_records - len(all_records)
                if remaining <= 0:
                    break
                page_limit = min(limit_per_page, remaining)
            else:
                page_limit = limit_per_page

            # Fetch page
            page = await self.get_page(limit=page_limit, next_token=next_token, **kwargs)

            all_records.extend(page.records)

            # Check if we should continue
            if not page.has_more:
                break

            next_token = page.next_token

            # Check if we've hit the max
            if max_records is not None and len(all_records) >= max_records:
                break

        return all_records[:max_records] if max_records else all_records

    async def iterate(self, limit_per_page: int = 25, **kwargs: Any) -> AsyncIterator[T]:
        """
        Iterate over all records across all pages.

        This is memory-efficient for large datasets as it yields
        records one at a time rather than loading all into memory.

        Args:
            limit_per_page: Number of records per page (max 25)
            **kwargs: Additional parameters to pass to fetch_page

        Yields:
            Individual records
        """
        next_token = None

        while True:
            page = await self.get_page(limit=limit_per_page, next_token=next_token, **kwargs)

            for record in page.records:
                yield record

            if not page.has_more:
                break

            next_token = page.next_token

    async def iterate_pages(self, limit_per_page: int = 25, **kwargs: Any) -> AsyncIterator[PaginatedResponse[T]]:
        """
        Iterate over pages of results.

        Args:
            limit_per_page: Number of records per page (max 25)
            **kwargs: Additional parameters to pass to fetch_page

        Yields:
            Page responses
        """
        next_token = None

        while True:
            page = await self.get_page(limit=limit_per_page, next_token=next_token, **kwargs)

            yield page

            if not page.has_more:
                break

            next_token = page.next_token


def parse_pagination_params(params: dict[str, Any]) -> dict[str, Any]:
    """
    Parse and validate pagination parameters.

    Args:
        params: Dictionary of parameters

    Returns:
        Validated parameters

    Raises:
        ValueError: If parameters are invalid
    """
    validated = {}

    # Validate limit
    if "limit" in params:
        limit = params["limit"]
        if not isinstance(limit, int) or limit < 1 or limit > DEFAULT_PAGE_SIZE:
            raise ValueError("Limit must be an integer between 1 and 25")
        validated["limit"] = limit

    # Pass through other params
    for key in ["start", "end", "nextToken"]:
        if key in params:
            validated[key] = params[key]

    return validated



"""Unit tests for Whoop API v2 models.

Copyright (c) 2024 Felix Geilert
"""

from datetime import datetime, timezone
from uuid import UUID

from whoopy.models.models_v2 import (
    Cycle,
    CycleScore,
    Recovery,
    RecoveryScore,
    ScoreState,
    Sleep,
    SleepNeeded,
    SleepStageSummary,
    UserBasicProfile,
    UserBodyMeasurement,
    WorkoutScore,
    WorkoutV2,
    ZoneDurations,
)


class TestUserModels:
    """Test user-related models."""

    def test_user_basic_profile(self):
        """Test UserBasicProfile model."""
        profile = UserBasicProfile(user_id=12345, email="test@example.com", first_name="John", last_name="Doe")

        assert profile.user_id == 12345
        assert profile.email == "test@example.com"
        assert profile.first_name == "John"
        assert profile.last_name == "Doe"

    def test_user_body_measurement(self):
        """Test UserBodyMeasurement model."""
        measurement = UserBodyMeasurement(height_meter=1.80, weight_kilogram=75.5, max_heart_rate=190)

        assert measurement.height_meter == 1.80
        assert measurement.weight_kilogram == 75.5
        assert measurement.max_heart_rate == 190


class TestCycleModels:
    """Test cycle-related models."""

    def test_cycle_score(self):
        """Test CycleScore model with computed calories."""
        score = CycleScore(strain=12.5, kilojoule=2000.0, average_heart_rate=120, max_heart_rate=180)

        assert score.strain == 12.5
        assert score.kilojoule == 2000.0
        assert score.average_heart_rate == 120
        assert score.max_heart_rate == 180
        # Test computed property
        assert abs(score.calories - 478.011) < 0.01

    def test_cycle_complete(self):
        """Test Cycle model with complete data."""
        now = datetime.now(timezone.utc)
        cycle = Cycle(
            id=12345,
            user_id=67890,
            created_at=now,
            updated_at=now,
            start=now,
            end=now,
            timezone_offset="-05:00",
            score_state=ScoreState.SCORED,
            score=CycleScore(strain=10.0, kilojoule=1500.0, average_heart_rate=110, max_heart_rate=170),
        )

        assert cycle.id == 12345
        assert cycle.user_id == 67890
        assert cycle.is_complete is True
        assert cycle.duration_hours == 0.0
        assert cycle.score is not None
        assert cycle.score.strain == 10.0

    def test_cycle_ongoing(self):
        """Test ongoing cycle without end time."""
        now = datetime.now(timezone.utc)
        cycle = Cycle(
            id=12345,
            user_id=67890,
            created_at=now,
            updated_at=now,
            start=now,
            end=None,
            timezone_offset="-05:00",
            score_state=ScoreState.PENDING_SCORE,
        )

        assert cycle.is_complete is False
        assert cycle.duration_hours is None


class TestSleepModels:
    """Test sleep-related models."""

    def test_sleep_stage_summary(self):
        """Test SleepStageSummary with computed properties."""
        summary = SleepStageSummary(
            total_in_bed_time_milli=28800000,  # 8 hours
            total_awake_time_milli=1800000,  # 30 minutes
            total_no_data_time_milli=0,
            total_light_sleep_time_milli=14400000,  # 4 hours
            total_slow_wave_sleep_time_milli=7200000,  # 2 hours
            total_rem_sleep_time_milli=5400000,  # 1.5 hours
            sleep_cycle_count=4,
            disturbance_count=3,
        )

        # Test computed properties
        assert summary.total_sleep_time_milli == 27000000  # 7.5 hours
        assert abs(summary.sleep_efficiency_percentage - 93.75) < 0.01

    def test_sleep_needed(self):
        """Test SleepNeeded with computed total."""
        needed = SleepNeeded(
            baseline_milli=28800000,  # 8 hours
            need_from_sleep_debt_milli=3600000,  # 1 hour
            need_from_recent_strain_milli=1800000,  # 30 minutes
            need_from_recent_nap_milli=-1800000,  # -30 minutes
        )

        assert needed.total_need_milli == 32400000  # 9 hours

    def test_sleep_complete(self):
        """Test Sleep model with UUID."""
        sleep_id = UUID("12345678-1234-5678-1234-567812345678")
        now = datetime.now(timezone.utc)
        sleep = Sleep(
            id=sleep_id,
            v1_id=12345,
            user_id=67890,
            created_at=now,
            updated_at=now,
            start=now,
            end=now,
            timezone_offset="-05:00",
            nap=False,
            score_state=ScoreState.SCORED,
        )

        assert sleep.id == sleep_id
        assert sleep.v1_id == 12345
        assert sleep.nap is False
        assert sleep.duration_hours == 0.0


class TestRecoveryModels:
    """Test recovery-related models."""

    def test_recovery_score(self):
        """Test RecoveryScore model."""
        score = RecoveryScore(
            user_calibrating=False,
            recovery_score=75.5,
            resting_heart_rate=55.0,
            hrv_rmssd_milli=45.5,
            spo2_percentage=98.5,
            skin_temp_celsius=33.5,
        )

        assert score.user_calibrating is False
        assert score.recovery_score == 75.5
        assert score.resting_heart_rate == 55.0
        assert score.hrv_rmssd_milli == 45.5
        assert score.spo2_percentage == 98.5
        assert score.skin_temp_celsius == 33.5

    def test_recovery(self):
        """Test Recovery model."""
        sleep_id = UUID("12345678-1234-5678-1234-567812345678")
        now = datetime.now(timezone.utc)
        recovery = Recovery(
            cycle_id=12345,
            sleep_id=sleep_id,
            user_id=67890,
            created_at=now,
            updated_at=now,
            score_state=ScoreState.SCORED,
            score=RecoveryScore(
                user_calibrating=False, recovery_score=80.0, resting_heart_rate=60.0, hrv_rmssd_milli=50.0
            ),
        )

        assert recovery.cycle_id == 12345
        assert recovery.sleep_id == sleep_id
        assert recovery.score.recovery_score == 80.0


class TestWorkoutModels:
    """Test workout-related models."""

    def test_zone_durations(self):
        """Test ZoneDurations with computed properties."""
        zones = ZoneDurations(
            zone_zero_milli=300000,
            zone_one_milli=600000,
            zone_two_milli=900000,
            zone_three_milli=900000,
            zone_four_milli=600000,
            zone_five_milli=300000,
        )

        assert zones.total_duration_milli == 3600000  # 1 hour

        percentages = zones.to_dict_percentage()
        assert abs(percentages["zone_zero_percentage"] - 8.33) < 0.01
        assert abs(percentages["zone_one_percentage"] - 16.67) < 0.01
        assert abs(percentages["zone_two_percentage"] - 25.0) < 0.01
        assert abs(percentages["zone_three_percentage"] - 25.0) < 0.01
        assert abs(percentages["zone_four_percentage"] - 16.67) < 0.01
        assert abs(percentages["zone_five_percentage"] - 8.33) < 0.01

    def test_workout_score(self):
        """Test WorkoutScore with computed calories."""
        score = WorkoutScore(
            strain=8.5,
            average_heart_rate=140,
            max_heart_rate=180,
            kilojoule=1000.0,
            percent_recorded=98.5,
            distance_meter=5000.0,
            altitude_gain_meter=100.0,
            altitude_change_meter=20.0,
            zone_durations=ZoneDurations(
                zone_zero_milli=300000,
                zone_one_milli=600000,
                zone_two_milli=900000,
                zone_three_milli=900000,
                zone_four_milli=600000,
                zone_five_milli=300000,
            ),
        )

        assert score.strain == 8.5
        assert abs(score.calories - 239.005) < 0.01
        assert score.distance_meter == 5000.0

    def test_workout_v2(self):
        """Test WorkoutV2 model with UUID."""
        workout_id = UUID("87654321-4321-8765-4321-876543218765")
        now = datetime.now(timezone.utc)
        workout = WorkoutV2(
            id=workout_id,
            v1_id=54321,
            sport_id=1,
            user_id=67890,
            created_at=now,
            updated_at=now,
            start=now,
            end=now,
            timezone_offset="-05:00",
            sport_name="running",
            score_state=ScoreState.SCORED,
        )

        assert workout.id == workout_id
        assert workout.v1_id == 54321
        assert workout.sport_id == 1
        assert workout.sport_name == "running"
        assert workout.duration_hours == 0.0


class TestEnums:
    """Test enum values."""

    def test_score_state_enum(self):
        """Test ScoreState enum values."""
        assert ScoreState.SCORED == "SCORED"
        assert ScoreState.PENDING_SCORE == "PENDING_SCORE"
        assert ScoreState.UNSCORABLE == "UNSCORABLE"



"""Unit tests for Whoop API v2 client.

Copyright (c) 2024 Felix Geilert
"""

import asyncio
from datetime import datetime, timezone
from unittest.mock import AsyncMock, Mock

import aiohttp
import pytest
from aioresponses import aioresponses

from whoopy.client_v2 import WhoopClientV2
from whoopy.exceptions import (
    RateLimitError,
    ResourceNotFoundError,
    ServerError,
    TokenExpiredError,
    ValidationError,
)
from whoopy.utils import RetryConfig, TokenInfo


class TestWhoopClientV2:
    """Test WhoopClientV2 functionality."""

    @pytest.fixture
    def token_info(self):
        """Create test token info."""
        return TokenInfo(
            access_token="test_access_token",
            expires_in=3600,
            refresh_token="test_refresh_token",
            scopes=["read:cycles", "read:sleep", "offline"],
        )

    @pytest.fixture
    def client(self, token_info):
        """Create test client."""
        return WhoopClientV2(token_info=token_info, client_id="test_client_id", client_secret="test_client_secret")

    @pytest.mark.asyncio
    async def test_client_context_manager(self, client):
        """Test client context manager creates and closes session."""
        async with client as whoop:
            assert whoop._session is not None
            assert isinstance(whoop._session, aiohttp.ClientSession)
            assert whoop._retry_session is not None
            assert whoop.cycles is not None
            assert whoop.sleep is not None
            assert whoop.recovery is not None
            assert whoop.workouts is not None
            assert whoop.user is not None

        # After exiting, session should be closed
        assert client._session.closed

    @pytest.mark.asyncio
    async def test_client_headers(self, client, token_info):
        """Test client sets proper headers."""
        async with client as whoop:
            headers = whoop.session.headers
            assert headers["Authorization"] == f"Bearer {token_info.access_token}"
            assert "User-Agent" in headers
            assert headers["Accept"] == "application/json"

    @pytest.mark.asyncio
    async def test_check_response_success(self, client):
        """Test successful response check."""
        async with client as whoop:
            response = Mock()
            response.status = 200

            # Should not raise any exception
            await whoop.check_response(response)

    @pytest.mark.asyncio
    async def test_check_response_errors(self, client):
        """Test error response handling."""
        async with client as whoop:
            # 400 Bad Request
            response = Mock()
            response.status = 400
            response.json = AsyncMock(return_value={"error": "bad request"})

            with pytest.raises(ValidationError):
                await whoop.check_response(response)

            # 401 Unauthorized
            response.status = 401
            with pytest.raises(TokenExpiredError):
                await whoop.check_response(response)

            # 404 Not Found
            response.status = 404
            with pytest.raises(ResourceNotFoundError):
                await whoop.check_response(response)

            # 429 Rate Limited
            response.status = 429
            response.headers = {"Retry-After": "60"}
            with pytest.raises(RateLimitError) as exc_info:
                await whoop.check_response(response)
            assert exc_info.value.retry_after == 60

            # 500 Server Error
            response.status = 500
            response.headers = {}
            with pytest.raises(ServerError):
                await whoop.check_response(response)

    @pytest.mark.asyncio
    async def test_refresh_token(self, client):
        """Test token refresh functionality."""
        async with client as whoop:
            with aioresponses() as m:
                new_token_data = {
                    "access_token": "new_access_token",
                    "expires_in": 3600,
                    "refresh_token": "new_refresh_token",
                    "token_type": "Bearer",
                }

                m.post("https://api.prod.whoop.com/oauth/oauth2/token", payload=new_token_data)

                await whoop.refresh_token()

                assert whoop.token_info.access_token == "new_access_token"
                assert whoop.token_info.refresh_token == "new_refresh_token"
                assert whoop.session.headers["Authorization"] == "Bearer new_access_token"

    @pytest.mark.asyncio
    async def test_request_with_token_refresh(self, client):
        """Test automatic token refresh on 401."""
        async with client as whoop:
            with aioresponses() as m:
                # First request returns 401
                m.get("https://api.prod.whoop.com/developer/v2/test", status=401)

                # Token refresh
                m.post(
                    "https://api.prod.whoop.com/oauth/oauth2/token",
                    payload={
                        "access_token": "new_access_token",
                        "expires_in": 3600,
                        "refresh_token": "new_refresh_token",
                        "token_type": "Bearer",
                    },
                )

                # Retry request succeeds
                m.get("https://api.prod.whoop.com/developer/v2/test", payload={"data": "test"})

                response = await whoop.request("GET", "test")
                data = await response.json()
                assert data == {"data": "test"}
                assert whoop.token_info.access_token == "new_access_token"


class TestAuthenticationMethods:
    """Test authentication class methods."""

    @pytest.mark.asyncio
    async def test_from_token(self):
        """Test creating client from token."""
        client = WhoopClientV2.from_token(
            access_token="test_token",
            expires_in=3600,
            refresh_token="refresh_token",
            scopes=["read:cycles"],
            client_id="client_id",
            client_secret="client_secret",
        )

        assert client.token_info.access_token == "test_token"
        assert client.token_info.refresh_token == "refresh_token"
        assert client.client_id == "client_id"
        assert client.client_secret == "client_secret"

    @pytest.mark.asyncio
    async def test_from_config(self, tmp_path):
        """Test creating client from config files."""
        # Create config file
        config_file = tmp_path / "config.json"
        config_file.write_text("""{
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
            "redirect_uri": "http://localhost:1234"
        }""")

        # Create token file
        token_file = tmp_path / ".whoop_credentials.json"
        token_file.write_text(f"""{{
            "access_token": "saved_token",
            "expires_in": 3600,
            "refresh_token": "saved_refresh",
            "scopes": ["read:cycles"],
            "token_type": "Bearer",
            "created_at": "{datetime.now(timezone.utc).isoformat()}"
        }}""")

        client = WhoopClientV2.from_config(config_path=str(config_file), token_path=str(token_file))

        assert client.token_info.access_token == "saved_token"
        assert client.client_id == "test_client_id"
        assert client.client_secret == "test_client_secret"

    def test_save_token(self, tmp_path):
        """Test saving token to file."""
        token_info = TokenInfo(
            access_token="test_token", expires_in=3600, refresh_token="refresh_token", scopes=["read:cycles"]
        )

        client = WhoopClientV2(token_info=token_info)

        token_file = tmp_path / "token.json"
        client.save_token(str(token_file))

        assert token_file.exists()

        # Load and verify
        import json

        with open(token_file) as f:
            saved_data = json.load(f)

        assert saved_data["access_token"] == "test_token"
        assert saved_data["refresh_token"] == "refresh_token"
        assert saved_data["scopes"] == ["read:cycles"]


class TestRetryLogic:
    """Test retry functionality."""

    @pytest.mark.asyncio
    async def test_retry_on_server_error(self):
        """Test retry on 500 error."""
        token_info = TokenInfo(access_token="test_token", expires_in=3600, refresh_token=None, scopes=[])

        retry_config = RetryConfig(max_attempts=3, base_delay=0.1, jitter=False)

        client = WhoopClientV2(token_info=token_info, retry_config=retry_config, auto_refresh_token=False)

        async with client as whoop:
            with aioresponses() as m:
                # First two attempts fail
                m.get("https://api.prod.whoop.com/developer/v2/test", status=500)
                m.get("https://api.prod.whoop.com/developer/v2/test", status=500)
                # Third attempt succeeds
                m.get("https://api.prod.whoop.com/developer/v2/test", payload={"data": "success"})

                response = await whoop.request("GET", "test")
                data = await response.json()
                assert data == {"data": "success"}

    @pytest.mark.asyncio
    async def test_retry_with_rate_limit(self):
        """Test retry with rate limit header."""
        token_info = TokenInfo(access_token="test_token", expires_in=3600, refresh_token=None, scopes=[])

        retry_config = RetryConfig(max_attempts=2, base_delay=0.1, jitter=False)

        client = WhoopClientV2(token_info=token_info, retry_config=retry_config, auto_refresh_token=False)

        async with client as whoop:
            with aioresponses() as m:
                # First attempt rate limited
                m.get("https://api.prod.whoop.com/developer/v2/test", status=429, headers={"Retry-After": "1"})
                # Second attempt succeeds
                m.get("https://api.prod.whoop.com/developer/v2/test", payload={"data": "success"})

                start_time = asyncio.get_event_loop().time()
                response = await whoop.request("GET", "test")
                end_time = asyncio.get_event_loop().time()

                data = await response.json()
                assert data == {"data": "success"}
                # Should have waited at least 1 second
                assert end_time - start_time >= 0.9  # Allow small variance



"""Request throttling for rate limit prevention.

Copyright (c) 2024 Felix Geilert
"""

import asyncio
import time
from typing import Any


class RequestThrottler:
    """Throttle requests to prevent rate limiting."""

    def __init__(self, delay: float = 0.0, max_concurrent: int = 10):
        """
        Initialize the throttler.

        Args:
            delay: Minimum delay in seconds between requests
            max_concurrent: Maximum number of concurrent requests
        """
        self.delay = delay
        self.max_concurrent = max_concurrent
        self._last_request_time = 0.0
        self._semaphore = asyncio.Semaphore(max_concurrent)
        self._lock = asyncio.Lock()

    async def acquire(self) -> None:
        """Acquire permission to make a request."""
        async with self._semaphore:
            if self.delay > 0:
                async with self._lock:
                    # Calculate time to wait
                    current_time = time.time()
                    time_since_last = current_time - self._last_request_time
                    wait_time = max(0, self.delay - time_since_last)

                    if wait_time > 0:
                        await asyncio.sleep(wait_time)

                    self._last_request_time = time.time()

    def __enter__(self) -> "RequestThrottler":
        """Context manager entry."""
        return self

    def __exit__(self, *args: Any) -> None:
        """Context manager exit."""

    async def __aenter__(self) -> "RequestThrottler":
        """Async context manager entry."""
        await self.acquire()
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Async context manager exit."""

    def adjust_delay(self, factor: float = 2.0) -> None:
        """
        Adjust the delay between requests.

        Args:
            factor: Multiplication factor for delay adjustment
        """
        self.delay = min(self.delay * factor, 5.0)  # Cap at 5 seconds



"""Exception hierarchy for Whoop API v2.

This module provides a comprehensive exception hierarchy for handling
various error conditions that can occur when interacting with the Whoop API.

Copyright (c) 2024 Felix Geilert
"""

from typing import Any


class WhoopException(Exception):
    """Base exception for all Whoop API errors."""

    def __init__(self, message: str, details: dict[str, Any] | None = None):
        super().__init__(message)
        self.message = message
        self.details = details or {}


class AuthenticationError(WhoopException):
    """Raised when authentication fails or credentials are invalid."""

    def __init__(self, message: str = "Authentication failed", details: dict[str, Any] | None = None):
        super().__init__(message, details)


class TokenExpiredError(AuthenticationError):
    """Raised when the access token has expired."""

    def __init__(self, message: str = "Access token has expired", details: dict[str, Any] | None = None):
        super().__init__(message, details)


class RefreshTokenError(AuthenticationError):
    """Raised when refreshing the access token fails."""

    def __init__(self, message: str = "Failed to refresh access token", details: dict[str, Any] | None = None):
        super().__init__(message, details)


class RateLimitError(WhoopException):
    """Raised when API rate limit is exceeded."""

    def __init__(
        self,
        message: str = "Rate limit exceeded",
        retry_after: int | None = None,
        details: dict[str, Any] | None = None,
    ):
        super().__init__(message, details)
        self.retry_after = retry_after


class ResourceNotFoundError(WhoopException):
    """Raised when a requested resource is not found (404)."""

    def __init__(
        self,
        resource_type: str,
        resource_id: str | None = None,
        message: str | None = None,
        details: dict[str, Any] | None = None,
    ):
        if message is None:
            if resource_id:
                message = f"{resource_type} with ID '{resource_id}' not found"
            else:
                message = f"{resource_type} not found"
        super().__init__(message, details)
        self.resource_type = resource_type
        self.resource_id = resource_id


class ValidationError(WhoopException):
    """Raised when request validation fails (400)."""

    def __init__(
        self,
        message: str = "Request validation failed",
        validation_errors: dict[str, Any] | None = None,
        details: dict[str, Any] | None = None,
    ):
        super().__init__(message, details)
        self.validation_errors = validation_errors or {}


class ServerError(WhoopException):
    """Raised when the server encounters an error (5xx)."""

    def __init__(
        self, status_code: int = 500, message: str = "Internal server error", details: dict[str, Any] | None = None
    ):
        super().__init__(message, details)
        self.status_code = status_code


class ConfigurationError(WhoopException):
    """Raised when there's a configuration issue with the client."""

    def __init__(self, message: str, details: dict[str, Any] | None = None):
        super().__init__(message, details)



"""Utility modules for Whoop API v2.

Copyright (c) 2024 Felix Geilert
"""

from .auth import OAuth2Helper, TokenInfo
from .pagination import PaginatedResponse, PaginationHelper
from .retry import RetryableSession, RetryConfig, retry_with_backoff
from .throttle import RequestThrottler

__all__ = [
    "OAuth2Helper",
    "PaginatedResponse",
    "PaginationHelper",
    "RequestThrottler",
    "RetryConfig",
    "RetryableSession",
    "TokenInfo",
    "retry_with_backoff",
]



"""Official Whoop API.

Generated against version 1.0 of the API.

Copyright 2022 (C) Felix Geilert
"""

import json
import os
import uuid
import webbrowser
from typing import Any

import requests
from typing_extensions import Self

from .constants import HTTP_OK, MIN_PASSWORD_LENGTH
from .handlers import handler_v1 as handlers

API_VERSION = "1"
API_BASE = "https://api.prod.whoop.com/"
API_AUTH = f"{API_BASE}oauth/oauth2"
SCOPES = [
    "offline",
    "read:recovery",
    "read:cycles",
    "read:sleep",
    "read:workout",
    "read:profile",
    "read:body_measurement",
]


class WhoopClient:
    def __init__(
        self,
        access_token: str,
        expires_in: int,
        scopes: list[str],
        refresh_token: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
    ):
        """Creates a new WhoopClient.

        Note that when client_id and client_secret are not provided, the refresh function will not work.

        Args:
            access_token (str): The access token.
            expires_in (int): The time until the token expires (in seconds).
            scopes (list[str]): The scopes that the token has.
            refresh_token (str, optional): The refresh token. Defaults to None.
            client_id (str, optional): The client ID. Defaults to None.
            client_secret (str, optional): The client secret. Defaults to None.
        """
        self.token = access_token
        self.expires_in = expires_in
        self.scopes = scopes
        self.refresh_token = refresh_token
        self._client_id = client_id
        self._client_secret = client_secret

        # create a session
        self.user_agent = "Python/3.X (X11; Linux x86_64)"
        self._update_session()

        # create a bunch of handlers
        self.user = handlers.WhoopUserHandler(self)
        self.cycle = handlers.WhoopCycleHandler(self)
        self.sleep = handlers.WhoopSleepHandler(self)
        self.workout = handlers.WhoopWorkoutHandler(self)
        self.recovery = handlers.WhoopRecoveryHandler(self)

    @property
    def _token(self) -> dict[str, Any]:
        return {
            "access_token": self.token,
            "expires_in": self.expires_in,
            "refresh_token": self.refresh_token,
            "scopes": self.scopes,
        }

    def _update_session(self) -> None:
        """Updates the session with the new token."""
        self._base_path = f"{API_BASE}developer/v1"
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Authorization": f"Bearer {self.token}",
                "User-Agent": self.user_agent,
            }
        )

    def store_token(self, path: str) -> None:
        """Stores the token to a file.

        Args:
            path (str): The path to the file (e.g. ".tokens/token.json").
        """
        # verify that folder exists
        base_dir = os.path.dirname(path)
        os.makedirs(base_dir, exist_ok=True)

        # store the token
        with open(path, "w") as f:
            json.dump(self._token, f)

    @classmethod
    def from_token(cls, path: str, client_id: str, client_secret: str, overwrite_token: bool = True) -> Self:
        """
        Loads a token from a file and creates a WhoopClient instance.

        Args:
            path (str): The path to the token file (e.g. ".tokens/token.json").
            client_id (str): The OAuth2 client ID.
            client_secret (str): The OAuth2 client secret.
            overwrite_token (bool): Whether to save the refreshed token back to file (default: True).

        Returns:
            WhoopClient: Authenticated client instance with refreshed token.
        """
        with open(path) as f:
            token = json.load(f)
        client = cls(
            token["access_token"],
            token["expires_in"],
            token["scopes"],
            token["refresh_token"],
            client_id,
            client_secret,
        )
        client.refresh()

        # check if token should be updated
        if overwrite_token is True:
            client.store_token(path)

        return client

    # retrieves the authorization url
    @classmethod
    def auth_url(
        cls,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        state: str | None = None,
        scopes: list[str] | None = None,
    ) -> tuple[str, str]:
        """
        Generates authorization URL for the Whoop OAuth2 flow.

        Args:
            client_id (str): The OAuth2 client ID.
            client_secret (str): The OAuth2 client secret.
            redirect_uri (str): The redirect URI registered with Whoop.
            state (str | None): Optional state parameter for CSRF protection (auto-generated if None).
            scopes (list[str] | None): OAuth2 scopes to request (uses all available if None).

        Returns:
            tuple[str, str]: A tuple of (authorization_url, state_parameter).

        Raises:
            ValueError: If state is provided but less than 8 characters.
        """
        # check state
        if not state:
            state = str(uuid.uuid4())

        # retrieve the data
        if len(state) < MIN_PASSWORD_LENGTH:
            raise ValueError("State must be at least 8 characters long")

        # generate the
        scopes = scopes or SCOPES
        scope = " ".join(scopes)
        res = (
            f"{API_AUTH}/auth?"
            f"scope={scope}&"
            f"client_id={client_id}&"
            f"client_secret={client_secret}&"
            f"state={state}&"
            f"redirect_uri={redirect_uri}&"
            "response_type=code"
        )

        return res, state

    @classmethod
    def _parse_token(cls, payload: dict[str, Any], scopes: list[str] | None) -> tuple[dict[str, Any], list[str]]:
        url = f"{API_AUTH}/token"

        # retrieve the codes
        res = requests.post(url, data=payload)
        if res.status_code != HTTP_OK:
            raise RuntimeError(f"Authorization failed with code {res.status_code}")
        codes = res.json()
        token_scopes = codes["scope"].split(" ")

        # validate scopes
        if scopes:
            for scope in scopes:
                if scope not in token_scopes:
                    raise ValueError(f"Scope {scope} not granted")

        return codes, token_scopes

    @classmethod
    def authorize(
        cls,
        code: str,
        client_id: str,
        client_secret: str,
        redirect_url: str = "https://jwt.ms/",
        scopes: list[str] | None = None,
    ) -> Self:
        """
        Authorize the client with the given authorization code.

        Exchanges an authorization code for access and refresh tokens.

        Args:
            code (str): Authorization code from OAuth2 callback.
            client_id (str): The OAuth2 client ID.
            client_secret (str): The OAuth2 client secret.
            redirect_url (str): The redirect URI used in authorization (default: "https://jwt.ms/").
            scopes (list[str] | None): Expected scopes to validate (raises error if not granted).

        Returns:
            WhoopClient: Authenticated client instance.

        Raises:
            RuntimeError: If authorization fails.
            ValueError: If requested scopes were not granted.
        """
        # generate request using the code
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_url,
        }
        codes, token_scopes = cls._parse_token(payload, scopes)

        # generate the client
        return cls(
            codes["access_token"],
            codes["expires_in"],
            token_scopes,
            codes.get("refresh_token", None),
            client_id=client_id,
            client_secret=client_secret,
        )

    @classmethod
    def auth_flow(
        cls,
        client_id: str,
        client_secret: str,
        redirect_url: str = "https://jwt.ms/",
        state: str | None = None,
        scopes: list[str] | None = None,
    ) -> Self:
        """Runs through the entire auth flow.

        Note: This requires to copy the code query attribute from the resulting redirect url.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.
            redirect_url (str, optional): The redirect URL. Defaults to "https://jwt.ms/".
            state (str, optional): The state passed through to the output of request. Defaults to None.
            scopes (list[str], optional): The scopes to request. Defaults to None.
                (In case of None, all scopes are requested.)

        Returns:
            The client object.
        """
        # retrieve url
        url, state = cls.auth_url(client_id, client_secret, redirect_url, state, scopes)

        # send user to auth and wait for code
        webbrowser.open(url)
        code = input("Copy Code Attribute from URL: ")

        # authorize
        return cls.authorize(code, client_id, client_secret, redirect_url, scopes)

        # complete

    def refresh(self) -> None:
        """
        Refreshes the access token using the refresh token.

        Updates the client's token information with new access token and expiry.

        Raises:
            ValueError: If no refresh token or client credentials are available.
            RuntimeError: If token refresh fails.
        """
        # verify client is setup correctly
        if self.refresh_token is None:
            raise ValueError("No refresh token provided")
        if self._client_id is None or self._client_secret is None:
            raise ValueError("No client id or secret provided")

        # generate request using the code
        payload = {
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
        }

        # retrieve the codes
        codes, _ = self._parse_token(payload, self.scopes)

        # update data
        self.token = codes["access_token"]
        self.expires_in = codes["expires_in"]
        self.refresh_token = codes.get("refresh_token", None)

        # update sess
        self._update_session()

    @classmethod
    def from_token_or_flow(cls, secret_json: str, token_path: str, scopes: list[str] | None = None) -> Self:
        """
        Creates a new WhoopClient from a token or by using the authorization flow.

        If a valid token exists at token_path, loads and refreshes it. Otherwise,
        initiates the OAuth2 authorization flow.

        Args:
            secret_json (str): Path to JSON file containing client_id, client_secret, and redirect_uri.
            token_path (str): Path to save/load the token file.
            scopes (list[str] | None): OAuth2 scopes to request (uses all available if None).

        Returns:
            WhoopClient: Authenticated client instance.
        """
        # load the secret
        with open(secret_json) as f:
            secret = json.load(f)

        # check if token exists
        if os.path.exists(token_path):
            return cls.from_token(
                token_path,
                secret["client_id"],
                secret["client_secret"],
                overwrite_token=True,
            )

        # create a new client
        client = cls.auth_flow(
            secret["client_id"],
            secret["client_secret"],
            secret["redirect_uri"],
            scopes=scopes,
        )
        client.store_token(token_path)

        return client



"""Base handler for Whoop API v2 resources.

Copyright (c) 2024 Felix Geilert
"""

from __future__ import annotations

from abc import ABC
from collections.abc import AsyncIterator
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    from whoopy.client_v2 import WhoopClientV2

import pandas as pd

from whoopy.exceptions import ResourceNotFoundError
from whoopy.models import models_v2 as models
from whoopy.utils import PaginatedResponse, PaginationHelper

T = TypeVar("T", bound=models.BaseWhoopModel)


class BaseHandler(ABC):  # noqa: B024
    """Base handler for API resources."""

    def __init__(self, client: WhoopClientV2):
        """
        Initialize handler.

        Args:
            client: WhoopClientV2 instance
        """
        self.client = client

    async def _get(self, path: str, **kwargs: Any) -> Any:
        """Make a GET request."""
        response = await self.client.request("GET", path, **kwargs)
        return await response.json()

    async def _post(self, path: str, **kwargs: Any) -> Any:
        """Make a POST request."""
        response = await self.client.request("POST", path, **kwargs)
        return await response.json()

    def _parse_datetime(self, dt: str | datetime | None) -> str | None:
        """Parse datetime to ISO format string."""
        if dt is None:
            return None

        if isinstance(dt, str):
            # Assume it's already in correct format
            return dt

        if isinstance(dt, datetime):
            # Ensure datetime is timezone-aware
            if dt.tzinfo is None:
                # Assume UTC for naive datetimes
                from datetime import timezone

                dt = dt.replace(tzinfo=timezone.utc)

            # Convert to ISO format with Z suffix
            return dt.isoformat().replace("+00:00", "Z")

        raise ValueError(f"Invalid datetime type: {type(dt)}")


class ResourceHandler(BaseHandler, Generic[T]):
    """Base handler for single resources."""

    def __init__(self, client: WhoopClientV2, resource_path: str, model_class: type[T]):
        """
        Initialize resource handler.

        Args:
            client: WhoopClientV2 instance
            resource_path: Base path for the resource
            model_class: Pydantic model class for the resource
        """
        BaseHandler.__init__(self, client)
        self.resource_path = resource_path
        self.model_class = model_class

    async def get_by_id(self, resource_id: str | int) -> T:
        """
        Get a single resource by ID.

        Args:
            resource_id: The resource ID

        Returns:
            The resource model instance

        Raises:
            ResourceNotFoundError: If resource not found
        """
        try:
            path = f"{self.resource_path}/{resource_id}"
            data = await self._get(path)
            return self.model_class(**data)
        except Exception as e:
            if "404" in str(e):
                raise ResourceNotFoundError(
                    resource_type=self.model_class.__name__, resource_id=str(resource_id)
                ) from e
            raise


class CollectionHandler(BaseHandler, Generic[T]):
    """Base handler for resource collections with pagination."""

    def __init__(
        self,
        client: WhoopClientV2,
        collection_path: str,
        model_class: type[T],
        response_class: type[PaginatedResponse],
    ):
        """
        Initialize collection handler.

        Args:
            client: WhoopClientV2 instance
            collection_path: Base path for the collection
            model_class: Pydantic model class for items
            response_class: Pydantic model class for paginated response
        """
        BaseHandler.__init__(self, client)
        self.collection_path = collection_path
        self.model_class = model_class
        self.response_class = response_class

        # Create pagination helper
        self._pagination = PaginationHelper(fetch_page=self._fetch_page, model_class=model_class)

    async def _fetch_page(self, **params: Any) -> PaginatedResponse[T]:
        """Fetch a single page of results."""
        # Clean up parameters
        cleaned_params = {}

        if "limit" in params and params["limit"] is not None:
            cleaned_params["limit"] = min(max(1, int(params["limit"])), 25)

        if "start" in params and params["start"] is not None:
            cleaned_params["start"] = self._parse_datetime(params["start"])  # type: ignore[assignment]

        if "end" in params and params["end"] is not None:
            cleaned_params["end"] = self._parse_datetime(params["end"])  # type: ignore[assignment]

        if "next_token" in params and params["next_token"] is not None:
            cleaned_params["nextToken"] = params["next_token"]

        # Make request
        data = await self._get(self.collection_path, params=cleaned_params)

        # Parse response
        response = self.response_class(**data)
        return PaginatedResponse(records=response.records, next_token=response.next_token)

    async def get_page(
        self,
        limit: int = 10,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        next_token: str | None = None,
    ) -> PaginatedResponse[T]:
        """
        Get a single page of results.

        Args:
            limit: Number of items per page (1-25)
            start: Start time filter
            end: End time filter
            next_token: Token for next page

        Returns:
            Paginated response with items and next token
        """
        return await self._pagination.get_page(limit=limit, start=start, end=end, next_token=next_token)

    async def get_all(
        self,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        limit_per_page: int = 25,
        max_records: int | None = None,
    ) -> list[T]:
        """
        Get all items across all pages.

        Args:
            start: Start time filter
            end: End time filter
            limit_per_page: Items per page (1-25)
            max_records: Maximum total items to fetch

        Returns:
            List of all items
        """
        return await self._pagination.get_all(
            start=start, end=end, limit_per_page=limit_per_page, max_records=max_records
        )

    async def iterate(
        self, start: str | datetime | None = None, end: str | datetime | None = None, limit_per_page: int = 25
    ) -> AsyncIterator[T]:
        """
        Iterate over all items across all pages.

        Args:
            start: Start time filter
            end: End time filter
            limit_per_page: Items per page (1-25)

        Yields:
            Individual items
        """
        async for item in self._pagination.iterate(start=start, end=end, limit_per_page=limit_per_page):
            yield item

    async def get_dataframe(
        self,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        limit_per_page: int = 25,
        max_records: int | None = None,
    ) -> pd.DataFrame:
        """
        Get all items as a pandas DataFrame.

        Args:
            start: Start time filter
            end: End time filter
            limit_per_page: Items per page (1-25)
            max_records: Maximum total items to fetch

        Returns:
            DataFrame with all items
        """
        items = await self.get_all(start=start, end=end, limit_per_page=limit_per_page, max_records=max_records)

        return models.models_to_dataframe(items)  # type: ignore[arg-type]


class CombinedHandler(ResourceHandler[T], CollectionHandler[T]):
    """Handler that supports both single resource and collection operations."""

    def __init__(
        self,
        client: WhoopClientV2,
        resource_path: str,
        collection_path: str,
        model_class: type[T],
        response_class: type[PaginatedResponse[Any]],
    ):
        """
        Initialize combined handler.

        Args:
            client: WhoopClientV2 instance
            resource_path: Base path for single resources
            collection_path: Base path for collections
            model_class: Pydantic model class
            response_class: Pydantic model class for paginated response
        """
        # Initialize both base classes
        ResourceHandler.__init__(self, client, resource_path, model_class)
        CollectionHandler.__init__(self, client, collection_path, model_class, response_class)

        # Ensure we use the client from BaseHandler
        self.client = client



"""Synchronous wrapper for the async Whoop API v2 client.

This module provides a synchronous interface to the async WhoopClientV2,
making it easy to use in non-async code while maintaining the benefits
of the modern implementation.

Copyright (c) 2024 Felix Geilert
"""

import asyncio
import functools
import logging
import threading
from collections.abc import Callable
from datetime import datetime
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from whoopy.handlers import handlers_v2
from uuid import UUID

import pandas as pd

from .client_v2 import WhoopClientV2
from .models import models_v2 as models
from .utils import RetryConfig, TokenInfo

T = TypeVar("T")


class EventLoopThread:
    """Thread with its own event loop for running async code."""

    def __init__(self) -> None:
        self.loop: asyncio.AbstractEventLoop | None = None
        self.thread: threading.Thread | None = None
        self._started = threading.Event()

    def start(self) -> None:
        """Start the event loop thread."""
        if self.thread is not None:
            return

        def _run_loop() -> None:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            self._started.set()
            self.loop.run_forever()

        self.thread = threading.Thread(target=_run_loop, daemon=True)
        self.thread.start()
        self._started.wait()  # Wait for loop to be ready

    def run_coroutine(self, coro: Any) -> Any:
        """Run a coroutine in the event loop."""
        if self.loop is None:
            self.start()

        if self.loop is None:
            raise RuntimeError("Event loop not initialized")
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)
        return future.result()

    def stop(self) -> None:
        """Stop the event loop and thread."""
        if self.loop is not None:
            self.loop.call_soon_threadsafe(self.loop.stop)
        if self.thread is not None:
            self.thread.join(timeout=1)
            self.thread = None
            self.loop = None


def async_to_sync(method: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to convert async methods to sync.

    Args:
        method: The async method to wrap

    Returns:
        A synchronous wrapper function
    """

    @functools.wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        # Ensure we have an event loop thread
        if not hasattr(self, "_loop_thread") or self._loop_thread is None:
            raise RuntimeError("Event loop thread not initialized")

        coro = method(self, *args, **kwargs)
        return self._loop_thread.run_coroutine(coro)

    return wrapper


class SyncUserHandler:
    """Synchronous wrapper for UserHandler."""

    def __init__(self, async_handler: "handlers_v2.UserHandler", loop_thread: EventLoopThread) -> None:
        """
        Initialize SyncUserHandler.

        Args:
            async_handler: The async UserHandler to wrap
            loop_thread: Event loop thread for running async code
        """
        self._handler = async_handler
        self._loop_thread = loop_thread

    @async_to_sync
    async def get_profile(self) -> models.UserBasicProfile:
        """Get the authenticated user's basic profile."""
        return await self._handler.get_profile()

    @async_to_sync
    async def get_body_measurements(self) -> models.UserBodyMeasurement:
        """Get the authenticated user's body measurements."""
        return await self._handler.get_body_measurements()


class SyncCollectionMixin:
    """Mixin for synchronous collection operations."""

    _handler: Any  # Will be set by subclasses
    _loop_thread: EventLoopThread  # Will be set by subclasses

    @async_to_sync
    async def get_page(
        self,
        limit: int = 10,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        next_token: str | None = None,
    ) -> Any:
        """
        Get a single page of results.

        Args:
            limit: Maximum number of items per page (default: 10)
            start: Start datetime for filtering results
            end: End datetime for filtering results
            next_token: Token for pagination to get next page

        Returns:
            Paginated response containing items and next_token
        """
        return await self._handler.get_page(limit=limit, start=start, end=end, next_token=next_token)

    @async_to_sync
    async def get_all(
        self,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        limit_per_page: int = 25,
        max_records: int | None = None,
    ) -> list[Any]:
        """
        Get all items across all pages.

        Args:
            start: Start datetime for filtering results
            end: End datetime for filtering results
            limit_per_page: Items per page (default: 25)
            max_records: Maximum total records to fetch (None for all)

        Returns:
            List of all items across all pages
        """
        return await self._handler.get_all(start=start, end=end, limit_per_page=limit_per_page, max_records=max_records)

    def iterate(
        self, start: str | datetime | None = None, end: str | datetime | None = None, limit_per_page: int = 25
    ) -> list[Any]:
        """
        Iterate over all items across all pages.

        Note: In sync mode, this returns all items as a list rather than yielding.

        Args:
            start: Start datetime for filtering results
            end: End datetime for filtering results
            limit_per_page: Items per page (default: 25)

        Returns:
            List of all items
        """

        async def _iterate() -> list[Any]:
            items = []
            async for item in self._handler.iterate(start=start, end=end, limit_per_page=limit_per_page):
                items.append(item)
            return items

        # Return all items as a list since we can't yield from sync
        return self._loop_thread.run_coroutine(_iterate())

    @async_to_sync
    async def get_dataframe(
        self,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        limit_per_page: int = 25,
        max_records: int | None = None,
    ) -> pd.DataFrame:
        # type: ignore[misc]
        """
        Get all items as a pandas DataFrame.

        Args:
            start: Start datetime for filtering results
            end: End datetime for filtering results
            limit_per_page: Items per page (default: 25)
            max_records: Maximum total records to fetch (None for all)

        Returns:
            DataFrame containing all items with flattened structure
        """
        return await self._handler.get_dataframe(
            start=start, end=end, limit_per_page=limit_per_page, max_records=max_records
        )


class SyncCycleHandler(SyncCollectionMixin):
    """Synchronous wrapper for CycleHandler."""

    def __init__(self, async_handler: Any, loop_thread: EventLoopThread) -> None:
        """
        Initialize SyncCycleHandler.

        Args:
            async_handler: The async CycleHandler to wrap
            loop_thread: Event loop thread for running async code
        """
        self._handler = async_handler
        self._loop_thread = loop_thread

    @async_to_sync
    async def get_by_id(self, cycle_id: int) -> models.Cycle:
        """
        Get a cycle by ID.

        Args:
            cycle_id: The unique identifier of the cycle

        Returns:
            The requested Cycle object

        Raises:
            ResourceNotFoundError: If cycle not found
        """
        return await self._handler.get_by_id(cycle_id)

    @async_to_sync
    async def get_sleep(self, cycle_id: int) -> models.Sleep:
        """
        Get the sleep for a specific cycle.

        Args:
            cycle_id: The unique identifier of the cycle

        Returns:
            Sleep data associated with the cycle

        Raises:
            ResourceNotFoundError: If cycle or sleep not found
        """
        return await self._handler.get_sleep(cycle_id)


class SyncSleepHandler(SyncCollectionMixin):
    """Synchronous wrapper for SleepHandler."""

    def __init__(self, async_handler: Any, loop_thread: EventLoopThread) -> None:
        """
        Initialize SyncSleepHandler.

        Args:
            async_handler: The async SleepHandler to wrap
            loop_thread: Event loop thread for running async code
        """
        self._handler = async_handler
        self._loop_thread = loop_thread

    @async_to_sync
    async def get_by_id(self, sleep_id: str | UUID) -> models.Sleep:
        """
        Get a sleep activity by ID.

        Args:
            sleep_id: The UUID of the sleep activity (as string or UUID object)

        Returns:
            The requested Sleep object

        Raises:
            ResourceNotFoundError: If sleep not found
        """
        return await self._handler.get_by_id(sleep_id)


class SyncRecoveryHandler(SyncCollectionMixin):
    """Synchronous wrapper for RecoveryHandler."""

    def __init__(self, async_handler: Any, loop_thread: EventLoopThread) -> None:
        """
        Initialize SyncRecoveryHandler.

        Args:
            async_handler: The async RecoveryHandler to wrap
            loop_thread: Event loop thread for running async code
        """
        self._handler = async_handler
        self._loop_thread = loop_thread

    @async_to_sync
    async def get_for_cycle(self, cycle_id: int) -> models.Recovery:
        """
        Get the recovery for a specific cycle.

        Args:
            cycle_id: The unique identifier of the cycle

        Returns:
            Recovery data for the specified cycle

        Raises:
            ResourceNotFoundError: If recovery not found for cycle
        """
        return await self._handler.get_for_cycle(cycle_id)


class SyncWorkoutHandler(SyncCollectionMixin):
    """Synchronous wrapper for WorkoutHandler."""

    def __init__(self, async_handler: Any, loop_thread: EventLoopThread) -> None:
        """
        Initialize SyncWorkoutHandler.

        Args:
            async_handler: The async WorkoutHandler to wrap
            loop_thread: Event loop thread for running async code
        """
        self._handler = async_handler
        self._loop_thread = loop_thread

    @async_to_sync
    async def get_by_id(self, workout_id: str | UUID) -> models.WorkoutV2:
        """
        Get a workout by ID.

        Args:
            workout_id: The UUID of the workout (as string or UUID object)

        Returns:
            The requested WorkoutV2 object

        Raises:
            ResourceNotFoundError: If workout not found
        """
        return await self._handler.get_by_id(workout_id)

    @async_to_sync
    async def get_by_sport(
        self,
        sport_name: str,
        start: str | datetime | None = None,
        end: str | datetime | None = None,
        limit_per_page: int = 25,
        max_records: int | None = None,
    ) -> list[models.WorkoutV2]:
        """
        Get all workouts for a specific sport.

        Args:
            sport_name: Name of the sport to filter by
            start: Start datetime for filtering results
            end: End datetime for filtering results
            limit_per_page: Items per page (default: 25)
            max_records: Maximum total records to fetch (None for all)

        Returns:
            List of workouts matching the sport name
        """
        return await self._handler.get_by_sport(
            sport_name=sport_name, start=start, end=end, limit_per_page=limit_per_page, max_records=max_records
        )


class WhoopClientV2Sync:
    """Synchronous wrapper for WhoopClientV2."""

    def __init__(
        self,
        token_info: TokenInfo | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        redirect_uri: str = "http://localhost:1234",
        retry_config: RetryConfig | None = None,
        auto_refresh_token: bool = True,
        logger: logging.Logger | None = None,
        request_delay: float = 0.0,
        max_concurrent_requests: int = 10,
    ):
        """
        Initialize the synchronous Whoop API v2 client.

        Args:
            token_info: OAuth2 token information
            client_id: OAuth2 client ID (required for token refresh)
            client_secret: OAuth2 client secret (required for token refresh)
            redirect_uri: OAuth2 redirect URI
            retry_config: Configuration for retry behavior
            auto_refresh_token: Automatically refresh expired tokens
            logger: Logger instance (creates default if None)
            request_delay: Delay in seconds between requests (default: 0)
            max_concurrent_requests: Maximum concurrent requests (default: 10)
        """
        self._async_client = WhoopClientV2(
            token_info=token_info,
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            retry_config=retry_config,
            auto_refresh_token=auto_refresh_token,
            logger=logger,
            request_delay=request_delay,
            max_concurrent_requests=max_concurrent_requests,
        )

        # Initialize sync handlers
        self._user: SyncUserHandler | None = None
        self._cycles: SyncCycleHandler | None = None
        self._sleep: SyncSleepHandler | None = None
        self._recovery: SyncRecoveryHandler | None = None
        self._workouts: SyncWorkoutHandler | None = None

        # Event loop thread for running async code
        self._loop_thread: EventLoopThread | None = None

        # Initialize the client
        self._initialize()

    def _initialize(self) -> None:
        """Initialize the async client and create sync handlers."""
        # We'll initialize handlers lazily when first accessed
        self._initialized = False
        self._session_context: WhoopClientV2 | None = None

    def _ensure_initialized(self) -> None:
        """Ensure the async client is initialized with an active session."""
        if not self._initialized:
            # Check if we have authentication before initializing
            if self._async_client.token_info is None:
                from .exceptions import AuthenticationError

                raise AuthenticationError(
                    "No authentication token available. Please authenticate first using "
                    "WhoopClientV2Sync.auth_flow() or provide a valid token."
                )

            # Create event loop thread
            self._loop_thread = EventLoopThread()
            self._loop_thread.start()

            async def _init() -> None:
                # Enter the async context to create session
                self._session_context = await self._async_client.__aenter__()

                # Create sync wrappers for handlers
                assert self._loop_thread is not None  # Type guard
                self._user = SyncUserHandler(self._session_context.user, self._loop_thread)
                self._cycles = SyncCycleHandler(self._session_context.cycles, self._loop_thread)
                self._sleep = SyncSleepHandler(self._session_context.sleep, self._loop_thread)
                self._recovery = SyncRecoveryHandler(self._session_context.recovery, self._loop_thread)
                self._workouts = SyncWorkoutHandler(self._session_context.workouts, self._loop_thread)

                self._initialized = True

            self._loop_thread.run_coroutine(_init())

    @property
    def user(self) -> SyncUserHandler:
        """Get the user handler."""
        self._ensure_initialized()
        if self._user is None:
            raise RuntimeError("Client not initialized")
        return self._user

    @property
    def cycles(self) -> SyncCycleHandler:
        """Get the cycles handler."""
        self._ensure_initialized()
        if self._cycles is None:
            raise RuntimeError("Client not initialized")
        return self._cycles

    @property
    def sleep(self) -> SyncSleepHandler:
        """Get the sleep handler."""
        self._ensure_initialized()
        if self._sleep is None:
            raise RuntimeError("Client not initialized")
        return self._sleep

    @property
    def recovery(self) -> SyncRecoveryHandler:
        """Get the recovery handler."""
        self._ensure_initialized()
        if self._recovery is None:
            raise RuntimeError("Client not initialized")
        return self._recovery

    @property
    def workouts(self) -> SyncWorkoutHandler:
        """Get the workouts handler."""
        self._ensure_initialized()
        if self._workouts is None:
            raise RuntimeError("Client not initialized")
        return self._workouts

    @property
    def token_info(self) -> TokenInfo | None:
        """Get current token information."""
        return self._async_client.token_info

    def save_token(self, path: str = ".whoop_credentials.json") -> None:
        """
        Save current token to file.

        Args:
            path: Path to save the token file (default: ".whoop_credentials.json")

        Raises:
            ValueError: If no token to save
        """
        self._async_client.save_token(path)

    def __enter__(self) -> "WhoopClientV2Sync":
        """Enter context manager."""
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit context manager and clean up resources."""
        self.close()

    def close(self) -> None:
        """Close the client and clean up resources."""
        if self._initialized and self._session_context:

            async def _cleanup() -> None:
                await self._async_client.__aexit__(None, None, None)

            if self._loop_thread:
                self._loop_thread.run_coroutine(_cleanup())
                self._loop_thread.stop()
                self._loop_thread = None

            self._initialized = False
            self._session_context = None

    @async_to_sync
    async def refresh_token(self) -> None:
        """Refresh the access token."""
        self._ensure_initialized()
        await self._async_client.refresh_token()

    # Class methods for authentication
    @classmethod
    def auth_flow(
        cls,
        client_id: str,
        client_secret: str,
        redirect_uri: str = "http://localhost:1234",
        scopes: list[str] | None = None,
        open_browser: bool = True,
        request_delay: float = 0.0,
        max_concurrent_requests: int = 10,
    ) -> "WhoopClientV2Sync":
        """
        Perform OAuth2 authorization flow.

        Args:
            client_id: OAuth2 client ID
            client_secret: OAuth2 client secret
            redirect_uri: OAuth2 redirect URI
            scopes: List of scopes to request
            open_browser: Whether to open the authorization URL in browser
            request_delay: Delay in seconds between requests (default: 0)
            max_concurrent_requests: Maximum concurrent requests (default: 10)

        Returns:
            Authenticated WhoopClientV2Sync instance
        """

        async def _auth() -> TokenInfo:
            client = await WhoopClientV2.auth_flow(
                client_id=client_id,
                client_secret=client_secret,
                redirect_uri=redirect_uri,
                scopes=scopes,
                open_browser=open_browser,
            )
            token = client.token_info
            if token is None:
                raise RuntimeError("Failed to authenticate")
            return token

        # Use asyncio.run for class method since no instance exists yet
        token_info = asyncio.run(_auth())

        return cls(
            token_info=token_info,
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            request_delay=request_delay,
            max_concurrent_requests=max_concurrent_requests,
        )

    @classmethod
    def from_token(
        cls,
        access_token: str,
        expires_in: int = 3600,
        refresh_token: str | None = None,
        scopes: list[str] | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
    ) -> "WhoopClientV2Sync":
        """
        Create client from existing token.

        Args:
            access_token: OAuth2 access token
            expires_in: Token validity duration in seconds (default: 3600)
            refresh_token: OAuth2 refresh token for token renewal
            scopes: List of OAuth2 scopes granted to the token
            client_id: OAuth2 client ID (required for token refresh)
            client_secret: OAuth2 client secret (required for token refresh)

        Returns:
            Configured WhoopClientV2Sync instance
        """
        client = WhoopClientV2.from_token(
            access_token=access_token,
            expires_in=expires_in,
            refresh_token=refresh_token,
            scopes=scopes,
            client_id=client_id,
            client_secret=client_secret,
        )

        return cls(token_info=client.token_info, client_id=client_id, client_secret=client_secret)

    @classmethod
    def from_config(
        cls, config_path: str = "config.json", token_path: str = ".whoop_credentials.json"
    ) -> "WhoopClientV2Sync":
        """
        Create client from configuration files.

        Args:
            config_path: Path to JSON config file containing client credentials
            token_path: Path to save/load token file

        Returns:
            Configured WhoopClientV2Sync instance

        Raises:
            ConfigurationError: If config file not found or invalid
        """
        client = WhoopClientV2.from_config(config_path=config_path, token_path=token_path)

        return cls(
            token_info=client.token_info,
            client_id=client.client_id,
            client_secret=client.client_secret,
            redirect_uri=client.redirect_uri,
        )
